/*
** Daedalus (Version 2.5) File: color.h
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains definitions for general graphics routines, that operate
** on a color bitmap, unrelated to Mazes.
**
** Old last code change: 6/29/1990.
** Last code change: 10/10/2013.
*/

/*
******************************************************************************
** Constants
******************************************************************************
*/

#define xColmap ((0x7FFFFFFF - 3) / 3)
#define yColmap 0x7FFFFFFF
#define cColorMain 16
#define cColor     29

#define kvBlack   Rgb(  0,   0,   0)
#define kvMaroon  Rgb(127,   0,   0)
#define kvDkGreen Rgb(  0, 127,   0)
#define kvMaize   Rgb(127, 127,   0)
#define kvDkBlue  Rgb(  0,   0, 127)
#define kvPurple  Rgb(127,   0, 127)
#define kvDkCyan  Rgb(  0, 127, 127)
#define kvLtGray  Rgb(191, 191, 191)
#define kvGray    Rgb(127, 127, 127)
#define kvRed     Rgb(255,   0,   0)
#define kvGreen   Rgb(  0, 255,   0)
#define kvYellow  Rgb(255, 255,   0)
#define kvBlue    Rgb(  0,   0, 255)
#define kvMagenta Rgb(255,   0, 255)
#define kvCyan    Rgb(  0, 255, 255)
#define kvWhite   Rgb(255, 255, 255)
#define kvDkGray  Rgb( 63,  63,  63)
#define kvOrange  Rgb(255, 127,   0)
#define kvPink    Rgb(255, 127, 127)
#define kvBrown   Rgb(127,  63,   0)

#define nHueMax  3600
#define nHueHalf 1800
#define nHue13   1200
#define nHue23   2400
#define nHue16   600
#define rHueMax  3600.0

enum {
  crpRing    = 1,
  crpSpoke   = 2,
  crpFade    = 3,
  crpY       = 4,
  crpX       = 5,
  crpDiagUL  = 6,
  crpDiagUR  = 7,
  crpSquare  = 8,
  crpDiamond = 9,
  crpSpiral  = 10,
  crpRing2   = 11,
  crpRings   = 12,
  crpSpoke2  = 13,
  crpRainbow = 14,
  crpRandom  = 15,
  crpRandom2 = 16,
  crpUnknown = 17,
};


/*
******************************************************************************
** Macros
******************************************************************************
*/

#define cbPixelC 3
#define DeallocateColmap(c) DeallocateBitmap(c)
#define CbColmapRow(x) ((x)*cbPixelC + 3 & ~3)
#define CbColmap(x, y) LMul(y, CbColmapRow(x))
#define CX(c) MX(c)
#define CY(c) MY(c)
#define CI(c, x, y) (((long)(y)*((c)->clRow) << 2) + (x)*cbPixelC)
#define CJ(c, i) ((byte *)(c) + cbBH + (i))
#define CK(pb) (((long)*(pb) << 16) | ((long)*((pb)+1) << 8) | *((pb)+2))
#define CL(pb, r, g, b) *(pb) = (b); *((pb)+1) = (g); *((pb)+2) = (r)
#define CM(pb, r, g, b) b = *(pb); g = *((pb)+1); r = *((pb)+2)
#define CG(c, x, y) _CG(c, x, y)
#define CP(c, x, y, kv) _CP(c, x, y, kv)
#define FLegalC(c, x, y) \
  ((x) >= 0 && (y) >= 0 && (x) < CX(c) && (y) < CY(c))

#define cTexture 64
#define NWSE(n, w, s, e) ((n) | (w) << 6 | (long)(s) << 12 | (long)(e) << 18)
#define UD(u, d) ((u) | (d) << 12)
#define UdU(l) ((int)((dword)(l) & 4095))
#define UdD(l) ((int)((dword)(l) >> 12))
#define ITextureWall(kv, x, y, d) \
  ((int)(((kv) >> ((d) * 6)) & (cTexture-1)))
#define CGTexture(c, x, y, d) ITextureWall(CG(c, x, y), x, y, d)
#define CPTexture(c, x, y, d, t) (Cg(c, x, y) & \
  ~((KV)(cTexture-1) << ((d) * 6)) | ((KV)(t) << ((d) * 6)))


/*
******************************************************************************
** Types
******************************************************************************
*/

typedef bitmap colmap;
typedef constbitmap constcolmap;

typedef struct _colorsettings {
  flag fRainbowBlend;
  int nRainbowStart;
  int nRainbowDistance;
  KV kvBlend1;
  KV kvBlend2;
  int nReplacePattern;
  flag fGraphNumber;
  long lGrayscale;
  flag fMandelbrot;
} CS;

extern CONST KV rgkv[cColor];
extern CONST char *rgszColor[cColor];
extern CS cs;


/*
******************************************************************************
** Color Primitives
******************************************************************************
*/

extern KV KvDialog(void);

extern KV _CG(constcolmap, int, int);
extern void _CP(colmap, int, int, KV);
extern KV Cg(constcolmap, int, int);
extern void Cp(colmap, int, int, KV);
extern flag FAllocateColmap(colmap *, int, int);
extern flag FColmapSizeSet(colmap *, int, int);
extern flag FColmapResizeTo(colmap *, int, int);
extern flag FColmapShiftBy(colmap *, int, int);
extern flag FColmapCopy(constcolmap, colmap *);
extern KV KvShade(KV, real);
extern KV KvBlend(KV, KV);
extern KV KvBlendR(KV, KV, real);
extern KV KvBlendN(KV, KV, int, int);
extern KV KvDiff(KV, KV);
extern KV Hue(int);
#define HueD(d) Hue((int)((d) * (nHueMax / nDegMax)))
#define HueR(r) Hue((int)((r) * (rHueMax / rDegMax)))
#define kvRandom Rgb(Rnd(0, 255), Rnd(0, 255), Rnd(0, 255))
#define kvRandomRainbow Hue(Rnd(0, nHueMax-1))
extern KV Hsl(int, int, int);


/*
******************************************************************************
** Color Bitmap Editing
******************************************************************************
*/

extern void BlockC(colmap, int, int, int, int, KV);
extern void LineXC(colmap, int, int, int, KV);
extern void LineYC(colmap, int, int, int, KV);
extern void BoxC(colmap, int, int, int, int, int, int, KV);
#define EdgeC(c, x1, y1, x2, y2, kv) BoxC(c, x1, y1, x2, y2, 1, 1, kv)
#define BoxAllC(c, xsiz, ysiz, kv) \
  BoxC(c, 0, 0, CX(c)-1, CY(c)-1, xsiz, ysiz, kv)
extern void BlockMoveC(constcolmap, colmap, int, int, int, int, int, int,
  int);
#define BlockPutC(c1, c2, x, y, nOp) \
  BlockMoveC(c1, c2, 0, 0, CX(c1)-1, CY(c1)-1, x, y, nOp);
extern void LineZC(colmap, int, int, int, int, KV);
extern void LineC(colmap, int, int, int, int, KV);
extern flag FTriangleC(colmap, int, int, int, int, int, int, KV, int);
extern flag FQuadrilateralC(colmap, int, int, int, int, int, int, int, int,
  KV, int);
extern void DiskC(colmap, int, int, int, int, KV);
extern void CircleC(colmap, int, int, int, int, KV);
extern void ArcPolygonC(colmap, int, int, int, int, int, int, int, int, KV);
#define PolygonC(c, x, y, rx, ry, d, kv) \
  ArcPolygonC(c, x, y, rx, ry, 0, d, 1, d, kv)
extern flag FFillC(colmap, int, int, KV, KV, flag);


/*
******************************************************************************
** Color Bitmap Transformations
******************************************************************************
*/

extern void ColmapSet(colmap, KV);
extern void ColmapGrayscale(colmap);
extern void ColmapBrightness(colmap, int, real, int);
extern void ColmapReplace(colmap, KV, KV, int, int, int, int);
extern void ColmapReplacePattern(colmap, KV, int, int, int, int, int);
extern void ColmapExchange(colmap, KV, KV);
#define ColmapReplaceSimple(c, kvFrom, kvTo) \
  ColmapReplace(c, kvFrom, kvTo, 0, 0, 0, 0)
#define ColmapReplacePatternSimple(c, kvFrom, nMode) \
  ColmapReplacePattern(c, kvFrom, nMode, 0, 0, 0, 0)
extern void ColmapOrAndKv(colmap, KV, int);
extern void ColmapOrAnd(colmap, constcolmap, int);
extern void ColmapAlpha(colmap, constcolmap, constcolmap);
extern void ColmapTessellate(colmap, constcolmap);
extern flag FColmapTessellateTo(colmap *, int, int);
#define FColmapTessellateBy(c, x, y) FColmapTessellateTo(c, \
  (int)((real)CX(*c)*(x) + rMinute), (int)((real)CY(*c)*(y) + rMinute))
extern void ColmapFlipX(colmap);
extern void ColmapFlipY(colmap);
extern flag FColmapTranspose(colmap *);
#define ColmapRotateR(pc) FColmapTranspose(pc); ColmapFlipX(*pc)
#define ColmapRotateL(pc) FColmapTranspose(pc); ColmapFlipY(*pc)
#define ColmapRotateA(c) ColmapFlipX(c); ColmapFlipY(c)
extern flag FColmapZoomTo(colmap *, int, int, flag);
#define FColmapZoomBy(c, x, y, f) FColmapZoomTo(c, \
  (int)((real)CX(*c)*(x) + rMinute), (int)((real)CY(*c)*(y) + rMinute), f)
extern flag FColmapBias(colmap *, int, int, int, int);
extern flag FColmapCollapse(colmap *);
extern flag FColmapBlur(colmap);
extern void ColmapContrast(colmap, flag);
extern flag FColmapTransform(colmap, int, int);
extern flag FZoomSprite(bitmap *, colmap *, constbitmap, constcolmap);
extern void ColmapStretchToEdge(colmap, constcolmap,
  int, int, int, int, int, int, int, int);
extern long ColmapCount(constcolmap);
extern long ColmapFind(constcolmap, int);
extern flag FColmapAccentBoundary(colmap *);
extern flag FColmapAccentContrast(colmap, flag);
extern flag FColmapSlide(colmap, int, int);
extern flag FColmapEditRowColumn(colmap *, int, int, flag, flag);
extern void DrawSeen(colmap, constbitmap, constcolmap, int, int, int, flag);
extern flag FColmapStereogram(colmap *, constbitmap, constcolmap, int, int);
extern long ColmapLifeGenerate(colmap, colmap *, flag);
extern long Evolution(colmap *, colmap *, int, flag, flag);
extern void Mandelbrot(colmap, real, real, real, real, int, flag);


/*
******************************************************************************
** Color Bitmap IO Routines
******************************************************************************
*/

extern flag FColmapGetFromBitmap(colmap *, constbitmap, KV, KV);
extern flag FColmapPutToBitmap(constcolmap, bitmap *, int);
extern void ColmapOrAndFromBitmap(colmap, constbitmap, KV, KV, int);
extern flag FColmapBlendFromBitmap(colmap *,
  constbitmap, constbitmap, constbitmap);
extern int ColmapGraphDistance(colmap *, constbitmap, constbitmap, KV, KV,
  int, int, flag);
extern long ColmapGraphDistance2(colmap *, constbitmap, KV, KV, flag);
extern flag FColmapAntialias(colmap *, constbitmap, KV, KV, int);
extern flag FReadColmapCore(colmap *, FILE *, int, int, int, int);
extern void WriteColmap(constcolmap, FILE *);
extern flag FReadColmapTarga(colmap *, FILE *);
extern void WriteColmapTarga(constcolmap, FILE *);
extern flag FReadColmapPaint(colmap *, FILE *);
extern flag FReadDaedalusColmapCore(colmap *, FILE *, int, int);
extern void WriteDaedalusColmap(constcolmap, FILE *, flag);
extern void WriteTextColmap(constcolmap, FILE *, flag);
extern int GetEditColor(char *);
extern void ConvertKvToSz(KV, char *, int);
extern KV ParseColor(char *, flag);
extern KV ReadColor(FILE *);

/* color.h */
