/*
** Daedalus (Version 2.5) File: daedalus.cpp
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains Daedalus specific routines not related to the
** underlying operating system.
**
** Created: 11/18/1993.
** Last code change: 10/10/2013.
*/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <time.h>
#include <math.h>

#include "resource.h"
#include "util.h"
#include "graphics.h"
#include "color.h"
#include "threed.h"
#include "maze.h"
#include "draw.h"
#include "daedalus.h"


// Global variables

WS ws = {fFalse, 0, 0, 0, 0, 0,
  0, 0, fTrue, fFalse, fFalse,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  fTrue, fFalse, fTrue, fTrue, fFalse, fFalse, fFalse, 0, 50,
  fFalse, fFalse, fFalse, fFalse,
  -1, fFalse,
  NULL, NULL,
    fFalse, fTrue, fFalse, fFalse, fFalse, fFalse, fFalse, fFalse, fFalse,
    0, 0, cmdCreatePerfect, cmdCreatePerfect, 0, fFalse, fFalse,
    NULL, NULL, NULL, 0, 0, 0, 0,
    NULL, 0, 0, 0, 0, NULL, NULL, 0, NULL, 0, NULL, NULL, 0, NULL, 0, NULL, 0,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
BM bm = {xStart, yStart, 0, 0, 0, 0, 0, 1, 1, 1,
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, 0L, NULL, 0L,
  fFalse, fFalse, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
DR dr = {0, 0, 0, 2, 0, 0, 0, 2,
  fFalse, fFalse, fTrue, 0, fFalse, fTrue, fTrue, fFalse, fFalse, fFalse, 0L,
  fFalse, fFalse, fTrue, fFalse, fFalse, fFalse,
  nEdgeTorus,
  nInsideSmooth,
  fTrue, fFalse, fFalse, fFalse, fFalse, fTrue, 20, 8, fTrue, fFalse, 500, 10,
  fTrue, 160, 20, 100, 0, -1, fTrue, 0.4, 0, 100, 55.0, nTransDefault,
  15, 15, 15, 20, 10, 8, 0, 0, 0, 0, fTrue,
  kvWhite, kvBlack, kvGray, kvRed, kvDkBlue, kvDkGreen, kvCyan, kvGray,
    kvBlack, kvMaize, kvLtGray, kvBrown, kvWhite, kvBlack, fFalse,
  fFalse, fFalse, fFalse,
  fFalse, fFalse, fFalse, kvGreen, kvYellow, Rgb(255, 255, 223),
    -1, -1, -1, -1, -1, -1, fFalse, 94001225, 0, 0, 50, 333, 10990099,
    fFalse, fFalse, fFalse, -1, fFalse, fFalse, fFalse, fFalse,
    -1, -1, -1, -1, -1, 0, -1, -1, 0,
  fFalse, 0.0, NULL, NULL, NULL, NULL, NULL, 50, 40};

// Constant data

CONST int rgcmdMouse[11] = {cmdMove7, cmdMove8, cmdMove9,
  cmdMove4, 0, cmdMove6, cmdMove1, cmdMove2, cmdMove3,
  cmdMoveU, cmdMoveD};
CONST PT rgptSize[cmdSizeLast - cmdSize01 + 1] =
  {{32, 16}, {32, 32}, {64, 32}, {64, 48}, {64, 64},
  {128, 64}, {128, 96}, {128, 128}, {128, 192}, {192, 192},
  {256, 192}, {256, 384}, {512, 384}, {512, 768}, {1024, 384},
  {1024, 768}, {1024, 1024}, {1024, 2048}, {2048, 4096}};

// Script names

CONST char *rgszScript[cmdScriptLast - cmdScript01 + 1] = {
  "demos.ds",    "wordmaze.ds", "gigamaze.ds", "maze4d.ds",   "maze5d.ds",
  "dragon.ds",   "pacman.ds",   "sokoban.ds",  "hunger.ds",
  "squared.ds",  "survmaz1.ds", "survmaz2.ds", "survmaz3.ds", "survmaz4.ds",
  "survmaz5.ds", "survmaz6.ds", "survmaz7.ds", "survmaz8.ds", "survmaz9.ds", 
  "carletn1.ds", "carletn2.ds", "stocker.ds",  "glacier.ds",  "safari.ds",
  "mousemaz.ds", "mandy.ds"};
CONST char *rgszShortcut[cmdScriptLast - cmdScript01 + 1] = {
  "Daedalus demos",        "Word Mazes",            "World's largest Maze",
  "4D Mazes",              "5D Mazes",              "Dragonslayer",
  "Pac-Man game",          "Sokoban game",          "The Hunger Games",
  "Survivor Squares game",
  "Survivor Maze game #1", "Survivor Maze game #2", "Survivor Maze game #3",
  "Survivor Maze game #4", "Survivor Maze game #5", "Survivor Maze game #6",
  "Survivor Maze game #7", "Survivor Maze game #8", "Survivor Maze game #9",
  "Carleton Farm Maze #1", "Carleton Farm Maze #2", "Stocker Farms Maze",
  "Glacier Maze game",     "Safari Maze",           "Mouse Maze game",
  "Mandelbrot set"};

// Command lines to run on startup. When embedding a complex script file, in
// it first delete the leading DS, then replace \ with \\, " with \", and ?
// with \?, to result in valid C++ strings.

CONST char *rgszStartup[] = {
  NULL}; // Must end with NULL


/*
******************************************************************************
** Daedalus Routines
******************************************************************************
*/

// Read a monochrome or color Windows bitmap from a file.

flag FReadBitmap(FILE *file, flag fNoHeader)
{
  int x, y, z, k;
  char sz[cchSzMax];

  if (!FReadBitmapHeader(file, fNoHeader, &x, &y, &z, &k))
    return fFalse;
  if (z != 1 && z != 4 && z != 8 && z != 16 && z != 24 && z != 32) {
    sprintf(S(sz), "This Windows bitmap has %d bits per pixel.\n"
      "Bitmaps must have 1, 4, 8, 16, 24, or 32 bits.\n", z);
    PrintSz_W(sz);
    return fFalse;
  }
  if (z == 1) {
    if (FReadBitmapCore(&bm.a, file, x, y))
      FShowColmap(fFalse);
  } else {
    if (FReadColmapCore(&bm.c, file, x, y, z, k))
      FShowColmap(fTrue);
  }
  return fTrue;
}


// Read a monochrome or color Daedalus bitmap from a file.

flag FReadDaedalusBitmap(FILE *file)
{
  int x, y, z;
  char ch, ch2;

  ch = getbyte(); ch2 = getbyte();
  if (ch != 'D' || ch2 != 'B') {
    PrintSz_W("This file does not look like a Daedalus bitmap.\n");
    return fFalse;
  }
  if (file != NULL)
    fscanf(file, "%d%d%d", &x, &y, &z);
  else {
    ws.iszStartup++;
    sscanf(ws.rgszStartup[ws.iszStartup], "%d%d%d", &x, &y, &z);
  }
  if (z != 1 && z != 24) {
    PrintSz_W(
      "This Daedalus bitmap is neither monochrome nor 24 bit color.\n");
    return fFalse;
  }
  if (file != NULL) {
    skipcrlf();
  } else {
    ws.iszStartup++;
  }
  if (z == 1) {
    if (FReadDaedalusBitmapCore(&bm.a, file, x, y))
      FShowColmap(fFalse);
  } else {
    if (FReadDaedalusColmapCore(&bm.c, file, x, y))
      FShowColmap(fTrue);
  }
  return fTrue;
}


// Resize the active bitmap. This implements the action of the Size dialog.

void DoSize(int x, int y, flag fShift, flag fClear)
{
  if (!bm.fColor) {
    if (!fShift) {
      if (x < 0)
        x = NAbs(MX(bm.a) + x);
      if (y < 0)
        y = NAbs(MY(bm.a) + y);
    }
    if (fClear) {
      if (fShift) {
        x += MX(bm.a); y += MY(bm.a);
      }
      if (FBitmapSizeSet(&bm.a, x, y))
        BitmapOff(bm.a);
    } else {
      if (!fShift)
        FBitmapResizeTo(&bm.a, x, y);
      else
        FBitmapShiftBy(&bm.a, x, y);
    }
    SetXyh(bm.a);
  } else {
    if (!fShift) {
      if (x < 0)
        x = NAbs(CX(bm.c) + x);
      if (y < 0)
        y = NAbs(CY(bm.c) + y);
    }
    if (fClear) {
      if (fShift) {
        x += CX(bm.c); y += CY(bm.c);
      }
      if (FColmapSizeSet(&bm.c, x, y))
        ColmapSet(bm.c, kvBlack);
    } else {
      if (!fShift)
        FColmapResizeTo(&bm.c, x, y);
      else
        FColmapShiftBy(&bm.c, x, y);
    }
  }
}


// Ensure there are at least the given number of slots available in the macro
// array, reallocating if needed.

flag FEnsureMacro(int cszNew)
{
  char **ppchT;

  if (cszNew <= ws.cszMacro)
    return fTrue;

  if (cszNew <= cMacro)
    cszNew = cMacro + 1;
  ppchT = (char **)ReallocateArray(ws.rgszMacro, ws.cszMacro,
    sizeof(char *), cszNew);
  if (ppchT == NULL)
    return fFalse;
  if (ws.rgszMacro != NULL)
    DeallocateP(ws.rgszMacro);
  ws.rgszMacro = ppchT;
  ws.cszMacro = cszNew;
  return fTrue;
}


// Ensure there are at least the given number of slots available in the custom
// variable array, reallocating if needed.

flag FEnsureLVar(int clNew)
{
  long *plT;

  if (clNew <= ws.clVar)
    return fTrue;

  if (clNew <= cLetter)
    clNew = cLetter + 1;
  plT = (long *)ReallocateArray(ws.rglVar, ws.clVar, sizeof(long), clNew);
  if (plT == NULL)
    return fFalse;
  if (ws.rglVar != NULL)
    DeallocateP(ws.rglVar);
  ws.rglVar = plT;
  ws.clVar = clNew;
  return fTrue;
}


// Ensure there are at least the given number of slots available in the custom
// string array, reallocating if needed.

flag FEnsureSzVar(int cszNew)
{
  char **ppchT;

  if (cszNew <= ws.cszVar)
    return fTrue;

  ppchT = (char **)ReallocateArray(ws.rgszVar, ws.cszVar, sizeof(char *),
    cszNew);
  if (ppchT == NULL)
    return fFalse;
  if (ws.rgszVar != NULL)
    DeallocateP(ws.rgszVar);
  ws.rgszVar = ppchT;
  ws.cszVar = cszNew;
  return fTrue;
}


// Set the given slot in the custom string array to the specified string.

flag SetSzVar(CONST char *sz, int cch, int isz)
{
  char *pch;

  if (!FEnsureSzVar(isz + 1))
    return fFalse;

  if (cch < 0)
    cch = CchSz(sz);
  pch = RgAllocate(cch + 1, char);
  if (pch == NULL)
    return fFalse;
  CopyRgb(sz, pch, cch);
  pch[cch] = chNull;

  if (ws.rgszVar[isz] != NULL)
    DeallocateP(ws.rgszVar[isz]);
  ws.rgszVar[isz] = pch;
  return fTrue;
}


// Return a pointer to the specified custom monochrome bitmap, as used for
// texture masks. This will point to NULL if it hasn't been allocated yet.

bitmap *BitmapGetMask(int ib)
{
  bitmap *pbT;

  // Special values allow referencing the standard bitmaps.
  Assert(ib >= -3);
  switch (ib) {
  case -1: return &bm.a;   // Bitmap -1 is the main bitmap
  case -2: return &bm.a2;  // Bitmap -2 is the temporary bitmap
  case -3: return &bm.a3;  // Bitmap -3 is the extra bitmap
  }

  // Ensure there are at least the given number of slots available in the
  // monochrome bitmap array, reallocating if needed.
  if (ib >= ws.cbMask) {
    pbT = (bitmap *)ReallocateArray(ws.rgbMask, ws.cbMask,
      sizeof(bitmap), ib+1);
    if (pbT == NULL)
      return NULL;
    if (ws.rgbMask != NULL)
      DeallocateP(ws.rgbMask);
    ws.rgbMask = pbT;
    ws.cbMask = ib+1;
  }
  return &ws.rgbMask[ib];
}


// Return a pointer to the specified custom color bitmap, as used for
// textures. This will point to NULL if it hasn't been allocated yet.

colmap *ColmapGetTexture(int ic)
{
  colmap *pcT;

  // Special values allow referencing the standard bitmaps.
  Assert(ic >= -5);
  switch (ic) {
  case -1: return &bm.c;   // Bitmap -1 is the main color bitmap
  case -2: return &bm.c2;  // Bitmap -2 is the temporary color bitmap
  case -3: return &bm.c3;  // Bitmap -3 is the extra color bitmap
  case -4: return &bm.cI;  // Bitmap -4 is the inside color bitmap
  case -5: return &bm.cR;  // Bitmap -5 is the inside rainbow bitmap
  }

  // Ensure there are at least the given number of slots available in the
  // color bitmap array, reallocating if needed.
  if (ic >= ws.ccTexture) {
    pcT = (colmap *)ReallocateArray(ws.rgcTexture, ws.ccTexture,
      sizeof(colmap), ic+1);
    if (pcT == NULL)
      return NULL;
    if (ws.rgcTexture != NULL)
      DeallocateP(ws.rgcTexture);
    ws.rgcTexture = pcT;
    ws.ccTexture = ic+1;
  }
  return &ws.rgcTexture[ic];
}


// Free all custom texture bitmaps and associated data. This is called from
// the Delete textures command, the ResetProgram operation, and when the
// program terminates.

void DeallocateTextures()
{
  int i;

  // Free each monochrome bitmap and the list itself.
  if (ws.rgbMask != NULL) {
    for (i = 0; i < ws.cbMask; i++)
      if (ws.rgbMask[i] != NULL)
        DeallocateBitmap(ws.rgbMask[i]);
    DeallocateP(ws.rgbMask);
    ws.rgbMask = NULL;
    ws.cbMask = 0;
  }

  // Free each color bitmap and the list itself.
  if (ws.rgcTexture != NULL) {
    for (i = 0; i < ws.ccTexture; i++)
      if (ws.rgcTexture[i] != NULL)
        DeallocateColmap(ws.rgcTexture[i]);
    DeallocateP(ws.rgcTexture);
    ws.rgcTexture = NULL;
    ws.ccTexture = 0;
  }
}


// Apply a texture map to the sky and ground background areas of the
// perspective inside view. Implements the Background apply texture command.

flag FTextureBackground()
{
  colmap *pc, c1 = bm.c;
  int iBase;
  flag fRet = fFalse;

  // Ensure at least one of the color and temporary color bitmaps exist.
  iBase = Max(ws.ccTexture, ws.cbMask); iBase = Max(iBase, 1);
  if (c1 == NULL && bm.c2 == NULL &&
    !FColmapGetFromBitmap(&c1, bm.a, dr.kvOff, dr.kvOn))
    return fFalse;
  if (ColmapGetTexture(iBase + 1) == NULL)
    goto LDone;

  // Set the sky/ground of inside view to a copy of the color bitmap, if
  // present. Put in next new texture slot to avoid conflicting with others.
  if (c1 != NULL) {
    pc = ColmapGetTexture(iBase);
    if (!FColmapCopy(c1, pc))
      goto LDone;
    dr.nTexture = iBase;
  }

  // Set the sky/ground of inside view for lower levels of 3D Mazes to a copy
  // of the temporary color bitmap, if present.
  if (bm.c2 != NULL) {
    pc = ColmapGetTexture(iBase + 1);
    if (!FColmapCopy(bm.c2, pc))
      goto LDone;
    dr.nTexture2 = iBase + 1;
  }

  fRet = fTrue;
LDone:
  if (c1 != NULL && c1 != bm.c)
    DeallocateColmap(c1);
  return fRet;
}


// Apply texture maps to all walls of perspective inside view. Implements the
// Color Walls, Overlay Walls, and Color Overlay Walls apply texture commands.

flag FTextureWall(int nMode, bit o)
{
  colmap c1 = bm.c, c2 = bm.c2, *pc, c;
  bitmap b1 = bm.a2, b2 = bm.a3, *pb;
  int iBase, n, x, y, n1, n2, n3;
  flag fShade = dr.rLightFactor != 0.0 && nMode != 2, fRet = fFalse;

  // Figure out how many different textures will be used. Narrow Walls on
  // doubles this number, also a non-zero Light Factor triples this number.
  iBase = Max(ws.ccTexture, ws.cbMask);
  n = (1 + dr.fNarrow) * (1 + fShade*2);
  if (ColmapGetTexture(iBase + n) == NULL ||
    BitmapGetMask(iBase + n) == NULL)
    return fFalse;

  // Set all pixels in the texture lookup bitmap to reference the textures
  // that are about to be defined.
  pc = ColmapGetTexture(iBase);
  if (!FColmapSizeSet(pc, MX(bm.a), MY(bm.a)))
    return fFalse;
  c = *pc;
  if (!dr.fNarrow) {

    // Narrow Walls off means all pixels have the same texture per side.
    ColmapSet(c, fShade ? NWSE(iBase + 2, iBase + 1, iBase + 3, iBase + 1) :
      NWSE(iBase + 1, iBase + 1, iBase + 1, iBase + 1));
  } else {

    // Narrow Walls on means narrow and normal walls have different textures.
    n1 = n2 = n3 = iBase + 2;
    if (fShade) {
      n2 = iBase + 4; n3 = iBase + 6;
    }
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++)
        CP(c, x, y,
          NWSE(n2 - FOdd(x), n1 - FOdd(y), n3 - FOdd(x), n1 - FOdd(y)));
  }

  // Define monochrome overlay textures. Applies to Overlay Walls and Color
  // Overlay Walls, but not Color Walls.
  if (nMode >= 2) {

    // If the overlays are to be white instead of black, set all pixels in the
    // overlay lookup bitmap to indicate all overlays are to be white.
    if (o) {
      pb = BitmapGetMask(0);
      if (!FBitmapSizeSet(pb, MX(bm.a), MY(bm.a)))
        return fFalse;
      BitmapSet(*pb, fOn);
    }

    // Make copies of bitmaps if they're going to be modified.
    if (nMode == 3) {
      b1 = b2 = NULL;
      if (bm.a2 != NULL && !FBitmapCopy(bm.a2, &b1))
        goto LDone;
      if (bm.a3 != NULL && !FBitmapCopy(bm.a3, &b2))
        goto LDone;
    }
    if (b1 == NULL && b2 == NULL && !FBitmapCopy(bm.a, &b1))
      goto LDone;
    if (nMode == 3) {
      if (b1 != NULL)
        BitmapReverse(b1);
      if (b2 != NULL)
        BitmapReverse(b2);
    }

    // If the temporary or extra bitmap doesn't exist, copy the other to it,
    // truncated or tessellated in proportion to the narrow walls setting.
    if (b1 == NULL) {
      if (!FAllocateBitmap(&b1,
        NMultDiv(MX(b2), dr.zCell, dr.zCellNarrow), MY(b2), b2))
        goto LDone;
      BitmapTessellate(b1, b2);
    } else if (dr.fNarrow && b2 == NULL) {
      if (!FAllocateBitmap(&b2,
        NMultDiv(MX(b1), dr.zCellNarrow, dr.zCell), MY(b1), b1))
        goto LDone;
      BitmapTessellate(b2, b1);
    }

    // Copy bitmaps to overlay textures.
    if (!FBitmapCopy(b1, BitmapGetMask(iBase + 1)))
      goto LDone;
    if (dr.fNarrow && !FBitmapCopy(b2, BitmapGetMask(iBase + 2)))
      goto LDone;
    if (fShade) {
      if (!FBitmapCopy(b1, BitmapGetMask(iBase + (dr.fNarrow ? 3 : 2))))
        goto LDone;
      if (!FBitmapCopy(b1, BitmapGetMask(iBase + (dr.fNarrow ? 5 : 3))))
        goto LDone;
      if (dr.fNarrow) {
        if (!FBitmapCopy(b2, BitmapGetMask(iBase + 4)))
          goto LDone;
        if (!FBitmapCopy(b2, BitmapGetMask(iBase + 6)))
          goto LDone;
      }
    }
  }

  // Define color textures. Applies to Color Walls and Color Overlay Walls,
  // but not Overlay Walls.
  if (nMode != 2) {

    // Make copies of color bitmaps if they're going to be modified.
    if (nMode == 3) {
      c1 = c2 = NULL;
      if (bm.c != NULL && !FColmapCopy(bm.c, &c1))
        goto LDone;
      if (bm.c2 != NULL && !FColmapCopy(bm.c2, &c2))
        goto LDone;
    }
    if (c1 == NULL && c2 == NULL &&
      !FColmapGetFromBitmap(&c1, bm.a, dr.kvOff, dr.kvOn))
      goto LDone;

    // If color or temporary color bitmap doesn't exist, copy the other to it,
    // truncated or tessellated in proportion to the narrow walls setting.
    if (c1 == NULL) {
      if (!FAllocateColmap(&c1,
        NMultDiv(CX(c2), dr.zCell, dr.zCellNarrow), CY(c2)))
        goto LDone;
      ColmapTessellate(c1, c2);
    } else if (dr.fNarrow && c2 == NULL) {
      if (!FAllocateColmap(&c2,
        NMultDiv(CX(c1), dr.zCellNarrow, dr.zCell), CY(c1)))
        goto LDone;
      ColmapTessellate(c2, c1);
    }
    if (nMode == 3) {
      ColmapOrAndFromBitmap(c1, b1, kvWhite, kvBlack, o);
      if (dr.fNarrow)
        ColmapOrAndFromBitmap(c2, b2, kvWhite, kvBlack, o);
    }

    // Copy color bitmaps to textures, shaded appropriately.
    if (!FColmapCopy(c1, ColmapGetTexture(iBase + 1)))
      goto LDone;
    if (dr.fNarrow && !FColmapCopy(c2, ColmapGetTexture(iBase + 2)))
      goto LDone;
    if (dr.rLightFactor != 0.0) {
      ColmapBrightness(c1, fFalse, -dr.rLightFactor, 0);
      if (nMode == 3 && o)
        ColmapOrAndFromBitmap(c1, b1, kvWhite, kvBlack, o);
      if (!FColmapCopy(c1, ColmapGetTexture(iBase + (dr.fNarrow ? 3 : 2))))
        goto LDone;
      FColmapCopy(*ColmapGetTexture(iBase + 1), &c1);
      ColmapBrightness(c1, fFalse, dr.rLightFactor, 0);
      if (nMode == 3 && !o)
        ColmapOrAndFromBitmap(c1, b1, kvWhite, kvBlack, o);
      if (!FColmapCopy(c1, ColmapGetTexture(iBase + (dr.fNarrow ? 5 : 3))))
        goto LDone;
      FColmapCopy(*ColmapGetTexture(iBase + 1), &c1);
      if (dr.fNarrow) {
        ColmapBrightness(c2, fFalse, -dr.rLightFactor, 0);
        if (nMode == 3 && o)
          ColmapOrAndFromBitmap(c2, b2, kvWhite, kvBlack, o);
        if (!FColmapCopy(c2, ColmapGetTexture(iBase + 4)))
          goto LDone;
        FColmapCopy(*ColmapGetTexture(iBase + 2), &c2);
        ColmapBrightness(c2, fFalse, dr.rLightFactor, 0);
        if (nMode == 3 && !o)
          ColmapOrAndFromBitmap(c2, b2, kvWhite, kvBlack, o);
        if (!FColmapCopy(c2, ColmapGetTexture(iBase + 6)))
          goto LDone;
        FColmapCopy(*ColmapGetTexture(iBase + 2), &c2);
      }
    }
  }

  dr.nTextureWall = iBase;
  fRet = fTrue;
LDone:
  // Free any temporary bitmaps that were allocated.
  if (c1 != NULL && c1 != bm.c)
    DeallocateColmap(c1);
  if (c2 != NULL && c2 != bm.c2)
    DeallocateColmap(c2);
  if (b1 != NULL && b1 != bm.a2)
    DeallocateBitmap(b1);
  if (b2 != NULL && b2 != bm.a3)
    DeallocateBitmap(b2);
  return fRet;
}


// Apply texture maps to all ground pixels of the perspective inside view.
// Implements the Color Ground, Overlay Ground, and Color Overlay Ground
// apply texture commands.

flag FTextureGround(flag fBlock, int nMode, bit o)
{
  colmap c1 = bm.c, c2 = NULL, c3 = NULL, c4 = NULL, *pc, c;
  bitmap b1 = bm.a2, b2 = NULL, b3 = NULL, b4 = NULL, *pb;
  int iBase, n, x, y;
  flag fRet = fFalse;

  // Figure out how many different textures will be used. Narrow Walls on
  // quadruples this number.
  iBase = Max(ws.ccTexture, ws.cbMask); iBase = Max(iBase, 1);
  n = (1 + dr.fNarrow * 3);
  if (ColmapGetTexture(iBase + n) == NULL ||
    BitmapGetMask(iBase + n) == NULL)
    return fFalse;

  // Set all pixels in the ground texture lookup bitmap to reference the
  // textures that are about to be defined.
  pc = ColmapGetTexture(iBase);
  if (!FColmapSizeSet(pc, MX(bm.a), MY(bm.a)))
    return fFalse;
  c = *pc;
  if (!dr.fNarrow) {

    // Narrow Walls off means all pixels have the same texture.
    n = iBase + 1;
    ColmapSet(c, UD(n, n));
  } else {

    // Narrow Walls on means narrow and normal cells have different textures.
    for (y = 0; y < CY(c); y++)
      for (x = 0; x < CX(c); x++) {
        n = iBase + 1 + (!FOdd(x) << 1) + !FOdd(y);
        CP(c, x, y, UD(n, n));
      }
  }

  // Define monochrome overlay textures. Applies to Overlay Ground and Color
  // Overlay Ground, but not Color Ground.
  if (nMode >= 2) {

    // If the overlays are to be white instead of black, set all pixels in the
    // overlay lookup bitmap to indicate all overlays are to be white.
    if (o) {
      pb = BitmapGetMask(iBase);
      if (!FBitmapSizeSet(pb, MX(bm.a), MY(bm.a)))
        return fFalse;
      BitmapSet(*pb, fOn);
    }

    // Make copies of bitmaps if they're going to be modified.
    if (nMode == 3) {
      b1 = NULL;
      if (bm.a2 != NULL && !FBitmapCopy(bm.a2, &b1))
        goto LDone;
    }
    if (b1 == NULL && !FBitmapCopy(bm.a, &b1))
      goto LDone;
    if (nMode == 3)
      BitmapReverse(b1);

    // If narrow walls is on, truncate or tessellate the first bitmap in
    // proportion to the narrow walls setting to create the others.
    if (dr.fNarrow) {
      if (!FAllocateBitmap(&b2,
        MX(b1), NMultDiv(MY(b1), dr.zCellNarrow, dr.zCell), b1))
        goto LDone;
      BitmapTessellate(b2, b1);
      if (!FAllocateBitmap(&b3,
        NMultDiv(MX(b1), dr.zCellNarrow, dr.zCell), MY(b1), b1))
        goto LDone;
      BitmapTessellate(b3, b1);
      if (!FAllocateBitmap(&b4, NMultDiv(MX(b1), dr.zCellNarrow, dr.zCell),
        NMultDiv(MY(b1), dr.zCellNarrow, dr.zCell), b1))
        goto LDone;
      BitmapTessellate(b4, b1);
    }

    // Copy bitmaps to overlay textures.
    if (!FBitmapCopy(b1, BitmapGetMask(iBase + 1)))
      goto LDone;
    if (dr.fNarrow && (!FBitmapCopy(b2, BitmapGetMask(iBase + 2)) ||
      !FBitmapCopy(b3, BitmapGetMask(iBase + 3)) ||
      !FBitmapCopy(b4, BitmapGetMask(iBase + 4))))
      goto LDone;
  }

  // Define color textures. Applies to Color Ground and Color Overlay Ground,
  // but not Overlay Ground.
  if (nMode != 2) {

    // Make copies of color bitmaps if they're going to be modified.
    if (c1 == NULL && !FColmapGetFromBitmap(&c1, bm.a, dr.kvOff, dr.kvOn))
      goto LDone;

    // If narrow walls is on, truncate or tessellate the first color bitmap in
    // proportion to the narrow walls setting to create the others.
    if (dr.fNarrow) {
      if (!FAllocateColmap(&c2,
        CX(c1), NMultDiv(CY(c1), dr.zCellNarrow, dr.zCell)))
        goto LDone;
      ColmapTessellate(c2, c1);
      if (!FAllocateColmap(&c3,
        NMultDiv(CX(c1), dr.zCellNarrow, dr.zCell), CY(c1)))
        goto LDone;
      ColmapTessellate(c3, c1);
      if (!FAllocateColmap(&c4, NMultDiv(CX(c1), dr.zCellNarrow, dr.zCell),
        NMultDiv(CY(c1), dr.zCellNarrow, dr.zCell)))
        goto LDone;
      ColmapTessellate(c4, c1);
    }
    if (nMode == 3)
      ColmapOrAndFromBitmap(c1, b1, kvWhite, kvBlack, o);

    // Copy color bitmap(s) to textures.
    if (!FColmapCopy(c1, ColmapGetTexture(iBase + 1)))
      goto LDone;
    if (dr.fNarrow && (!FColmapCopy(c2, ColmapGetTexture(iBase + 2)) ||
      !FColmapCopy(c3, ColmapGetTexture(iBase + 3)) ||
      !FColmapCopy(c4, ColmapGetTexture(iBase + 4))))
      goto LDone;
  }

  if (fBlock)
    dr.nTextureVar = iBase;
  else {
    if (!FBitmapSizeSet(&bm.a2, MX(bm.a), MY(bm.a)))
      goto LDone;
    BitmapOn(bm.a2);
    dr.nTextureDirt = iBase;
  }
  fRet = fTrue;
LDone:
  // Free any temporary bitmaps that were allocated.
  if (c1 != NULL && c1 != bm.c)
    DeallocateColmap(c1);
  if (c2 != NULL)
    DeallocateColmap(c2);
  if (c3 != NULL)
    DeallocateColmap(c3);
  if (c4 != NULL)
    DeallocateColmap(c4);
  if (b1 != NULL && b1 != bm.a2)
    DeallocateBitmap(b1);
  if (b2 != NULL)
    DeallocateBitmap(b2);
  if (b3 != NULL)
    DeallocateBitmap(b3);
  if (b4 != NULL)
    DeallocateBitmap(b4);
  return fRet;
}


// Make all cells in the perspective inside view have a ceiling marking.
// Implements the Ceiling apply texture command.

flag FTextureCeiling()
{
  bitmap *pb;
  colmap c, *pc;

  if (dr.nMarkSky < 0) {
    dr.nMarkSky = Max(ws.ccTexture, ws.cbMask);
    dr.nMarkSky = Max(dr.nMarkSky, 1);
  }
  pb = BitmapGetMask(dr.nMarkSky);
  if (pb == NULL || !FBitmapCopy(bm.a, pb))
    return fFalse;
  BitmapOn(*pb);

  // Make ceiling colors be a copy of the ground marking colors, if any.
  c = dr.fMarkColor ? bm.c2 : bm.c;
  if (c != NULL) {
    pc = ColmapGetTexture(dr.nMarkSky);
    if (pc == NULL || !FColmapCopy(c, pc))
      return fFalse;
  }
  return fTrue;
}


// Convert a Maze so it appears to be on hilly terrain in the perspective
// inside view. Implements the Variable Height Walls apply texture command.

flag FWallVariable(int nPattern)
{
  bitmap *pb;
  colmap c, *pc;
  int x, y, z, d, xnew, ynew, zHi, zLo;
  KV kv;
  flag f1, f2;

  if (dr.nWallVar < 0) {
    dr.nWallVar = Max(ws.ccTexture, ws.cbMask);
    dr.nWallVar = Max(dr.nWallVar, 1);
  }
  pb = BitmapGetMask(dr.nWallVar);
  if (pb == NULL || !FBitmapSizeSet(pb, MX(bm.a), MY(bm.a)))
    return fFalse;
  BitmapOn(*pb);

  pc = ColmapGetTexture(dr.nWallVar);
  if (pc == NULL || !FColmapSizeSet(pc, MX(bm.a), MY(bm.a)))
    return fFalse;
  c = *pc;

  // Compute the base altitude for each block.
  ColmapSet(c, kvBlack);
  f1 = cs.fGraphNumber; f2 = cs.fRainbowBlend;
  cs.fGraphNumber = fTrue; cs.fRainbowBlend = fFalse;
  ColmapReplacePattern(c, kvBlack, nPattern, dr.x, dr.y, dr.x2, dr.y2);
  cs.fGraphNumber = f1; cs.fRainbowBlend = f2;

  // Compute the color for each block.
  if (!FColmapSizeSet(&bm.c, MX(bm.a), MY(bm.a)))
    return fFalse;
  ColmapSet(bm.c, kvBlack);
  ColmapReplacePattern(bm.c, kvBlack, nPattern, dr.x, dr.y, dr.x2, dr.y2);

  // Determine the maximum possible bottom height of each block.
  for (y = 0; y < MY(bm.a); y++)
    for (x = 0; x < MX(bm.a); x++) {
      zLo = zHi = CG(c, x, y);
      for (d = 0; d < DIRS2; d++) {
        xnew = x + xoff[d]; ynew = y + yoff[d];
        z = UdU(Cg(c, xnew, ynew));
        zLo = Min(zLo, z);
      }
      CP(c, x, y, UD(zHi, zLo));
    }

  // Increase the top height of each block if a wall at that spot in the Maze.
  for (y = 0; y < MY(bm.a); y++)
    for (x = 0; x < MX(bm.a); x++) {
      kv = CG(c, x, y);
      zHi = UdU(kv) + 4;
      if (MG(bm.a, x, y))
        zHi += iVar;
      else
        CP(bm.c, x, y, KvBlend(CG(bm.c, x, y), GrayN(zHi * 255 / 3600)));
      CP(c, x, y, UD(zHi, UdD(kv)));
    }

  BitmapOn(bm.a);
  return fTrue;
}


// Convert a formatted range of characters to a string ready to display,
// translating character sequences as appropriate. This is used by the
// various Message and SetString operations.

void FormatRgchToSz(CONST char *pch, int cch, char *sz, int cchMax)
{
  CONST char *pchSrc = pch;
  char *pchDst = sz, ch1, ch2;
  long l;

  while (pchSrc - pch < cch && pchDst - sz < cchMax - 1) {
    ch1 = *pchSrc; ch2 = *(pchSrc + 1);
    if (ch1 == '\\') {
      if (ch2 == '\\')          // "\\" means a single "\"
        pchSrc++;
      else if (ch2 == 'n') {    // "\n" means a new line
        *pchDst++ = '\n';
        pchSrc += 2;
        continue;
      } else if (ch2 == 'r') {  // "\r" means a carriage return
        *pchDst++ = '\r';
        pchSrc += 2;
        continue;
      } else if (ch2 == 't') {  // "\t" means a tab
        *pchDst++ = chTab;
        pchSrc += 2;
        continue;
      }
    }
    if (ch1 == '@') {
      if (ch2 == '@')           // "@@" means a single "@"
        pchSrc++;
      else {

        // Expand "@x" to the number contained in that custom variable.
        // Expand "@n" to custom variable number n.
        ch2 = ChCap(ch2);
        if (FCapCh(ch2) || FDigitCh(ch2)) {
          if (FCapCh(ch2))
            l = ch2 - '@';
          else {
            l = ch2 - '0';
            loop {
              ch2 = *(pchSrc + 2);
              if (!FDigitCh(ch2))
                break;
              l = l * 10 + (ch2 - '0');
              pchSrc++;
            }
          }
          sprintf(SS(pchDst, cchMax - (pchDst - sz)), "%ld", LVar(l));
          while (*pchDst)
            pchDst++;
          pchSrc += 2;
          continue;
        }
      }
    }
    if (ch1 == '$') {
      if (ch2 == '$')           // "$$" means a single "$"
        pchSrc++;
      else {

        // Expand "$x" to the custom string pointed to by custom variable @x.
        // Expand "$n" to custom string number n.
        ch2 = ChCap(ch2);
        if (FCapCh(ch2) || FDigitCh(ch2)) {
          if (FCapCh(ch2))
            l = LVar(ch2 - '@');
          else {
            l = ch2 - '0';
            loop {
              ch2 = *(pchSrc + 2);
              if (!FDigitCh(ch2))
                break;
              l = l * 10 + (ch2 - '0');
              pchSrc++;
            }
          }
          if (l >= 0 && l < ws.cszVar && ws.rgszVar[l] != NULL) {
            CopyRgchToSz(ws.rgszVar[l], CchSz(ws.rgszVar[l]),
              pchDst, cchMax - PD(pchDst - sz));
            while (*pchDst)
              pchDst++;
          }
          pchSrc += 2;
          continue;
        }
      }
    }
    *pchDst++ = *pchSrc++;
  }
  *pchDst = chNull;
}


// Set pixels in the temporary bitmap, at the passed in coordinates and at the
// coordinates of the dot. This is called from the dot movement commands when
// Do Drag On Temp is set.

void DotTemp(int x, int y, int z)
{
  int x1, y1, x2, y2;

  if (bm.a2 == NULL)
    FBitmapCopy(bm.a, &bm.a2);
  if (bm.a2 != NULL) {
    if (dr.f3D) {
      x1 = Tx(bm.a, x, z); y1 = Ty(bm.a, y, z);
      x2 = Tx(bm.a, dr.x, dr.z); y2 = Ty(bm.a, dr.y, dr.z);
    } else {
      x1 = x; y1 = y;
      x2 = dr.x; y2 = dr.y;
    }
    Mp(bm.a2, x1, y1, dr.fSet);
    Mp(bm.a2, x2, y2, dr.fSet);
    if (dr.fNarrow && (!dr.f3D || z == dr.z))
      Mp(bm.a2, (x1 + x2) >> 1, (y1 + y2) >> 1, dr.fSet);
  }
}


// Adjust a bitmap either before or after a Maze has been created. This
// implements the scenarios when the Polished Mazes setting is active.
//
// Mode 0: Prepare to make a Maze. Zooms the bitmap down if the polish process
//   zooms it up, so successive Mazes in polish mode won't enlarge the bitmap.
// Mode 1: Polish a raw Maze that was just created in the bitmap.
// Mode 2: Leaving polish mode, so unpolish the Maze in the bitmap.

void PolishMaze(int wCmd, int nMode)
{
  bitmap bT;

  if (!dr.fPolishMaze)
    return;

  // Polishing most Maze types means creating a bias zoomed version of it.
  switch (wCmd) {
  case cmdCreatePerfect:
  case cmdCreateBraid:
  case cmdCreateUnicursal:
  case cmdCreateRecursive:
  case cmdCreatePrim:
  case cmdCreateKruskal:
  case cmdCreateAldous:
  case cmdCreateWilson:
  case cmdCreateEller:
  case cmdCreateTree:
  case cmdCreateDivision:
  case cmdCreateBinary:
  case cmdCreateSidewinder:
  case cmdCreateRandom:
  case cmdCreateSpiral:
  case cmdCreateDiagonal:
  case cmdCreateSegment:
  case cmdCreateFractal:
  case cmdCreateWeave:
  case cmdCreatePlanair:
  case cmdCreateInfinite0:
  case cmdCreateInfinite1:
  case cmdCreateInfinite2:
  case cmdCreateInfinite3:
    if (FBetween(wCmd, cmdCreateInfinite0, cmdCreateInfinite3)) {
      if (nMode == 0)
        nMode = 2;
    }
    switch (nMode) {
    case 0:
      FBitmapSizeSet(&bm.a, (MX(bm.a) + 3) >> 1, (MY(bm.a) + 3) >> 1);
      break;
    case 1:
      FBitmapResizeTo(&bm.a, Odd(MX(bm.a)), Odd(MY(bm.a)));
      FBitmapBias(&bm.a, 1, 3, 1, 3);
      dr.x <<= 1; dr.y <<= 1;
      break;
    case 2:
      FBitmapResizeTo(&bm.a, MX(bm.a) + 3, MY(bm.a) + 3);
      FBitmapZoomTo(&bm.a, MX(bm.a) >> 1, MY(bm.a) >> 1, fFalse);
      dr.x >>= 1; dr.y >>= 1;
      break;
    }
    break;

  // Recursive fractal Mazes are like the above, but affect color bitmap too.
  case cmdCreateFractal2:
    switch (nMode) {
    case 1:
      if (bm.c != NULL && CX(bm.c) == MX(bm.a) && CY(bm.c) == MY(bm.a))
        FColmapBias(&bm.c, 1, 3, 1, 3);
      BitmapReverse(bm.a);
      FBitmapBias(&bm.a, 1, 3, 1, 3);
      dr.x <<= 1; dr.y <<= 1;
      break;
    case 2:
      FBitmapResizeTo(&bm.a, MX(bm.a) + 3, MY(bm.a) + 3);
      FBitmapZoomTo(&bm.a, MX(bm.a) >> 1, MY(bm.a) >> 1, fFalse);
      dr.x >>= 1; dr.y >>= 1;
      FBitmapResizeTo(&bm.a, MX(bm.a) - 1, MY(bm.a) - 1);
      BitmapReverse(bm.a);
      break;
    }
    break;

  // For Mazes with non-orthogonal walls, smooth them out.
  case cmdCreateCrack:
  case cmdCreateCavern:
  case cmdCreateArrow:
  case cmdCreateTilt:
    switch (nMode) {
    case 0:
      FBitmapSizeSet(&bm.a, MX(bm.a) >> 1, MY(bm.a) >> 1);
      break;
    case 1:
      BitmapSmooth(&bm.a, fTrue);
      dr.x <<= 1; dr.y <<= 1;
      break;
    case 2:
      BitmapReverse(bm.a);
      FBitmapZoomTo(&bm.a, MX(bm.a) >> 1, MY(bm.a) >> 1, fTrue);
      BitmapReverse(bm.a);
      dr.x >>= 1; dr.y >>= 1;
      break;
    }
    break;

  // Polished Zeta Mazes are like above, but give bias to off pixels.
  case cmdCreateZeta:
    switch (nMode) {
    case 0:
      FBitmapSizeSet(&bm.a, (MX(bm.a) + 3) >> 1, (MY(bm.a) + 3) >> 1);
      break;
    case 1:
      FBitmapResizeTo(&bm.a, Odd(MX(bm.a)), Odd(MY(bm.a)));
      BitmapReverse(bm.a);
      BitmapSmooth(&bm.a, fTrue);
      BitmapReverse(bm.a);
      dr.x <<= 1; dr.y <<= 1;
      break;
    case 2:
      FBitmapZoomTo(&bm.a, MX(bm.a) >> 1, MY(bm.a) >> 1, fTrue);
      FBitmapResizeTo(&bm.a, MX(bm.a) + 1, MY(bm.a) + 1);
      dr.x >>= 1; dr.y >>= 1;
      break;
    }
    break;

  // For 3D Mazes, convert to a clarified version in main and color bitmaps.
  case cmdCreate3D:
    if (nMode == 1) {
      FColmapClarify3D(bm.a, &bm.c, dr.kvOff, dr.kvOn);
      ColmapReplaceSimple(bm.c, kvGreen,  Rgb(0,   170, 0));
      ColmapReplaceSimple(bm.c, kvYellow, Rgb(170, 170, 0));
      FClarify3D(&bm.a);
      ws.wCmdSpree = cmdPolishMaze;
    }
    break;

  // For 4D Mazes, convert to a clarified version in main and color bitmaps.
  case cmdCreate4D:
    if (nMode == 1) {
      FColmapClarify4D(bm.a, &bm.c, dr.kvOff, dr.kvOn);
      ColmapReplaceSimple(bm.c, kvGreen,  Rgb(0,   170, 0));
      ColmapReplaceSimple(bm.c, kvYellow, Rgb(170, 170, 0));
      FClarify4D(&bm.a);
      ws.wCmdSpree = cmdPolishMaze;
    }
    break;

  // For Hypermazes, draw a 3D overview of it.
  case cmdCreateHyper:
    if (nMode == 1) {
      DrawOverviewCube(&bm.a);
      BitmapReverse(bm.a);
      FBitmapCollapse(&bm.a);
      ws.wCmdSpree = cmdPolishMaze;
    }
    break;

  // For Chartres Replica, fill in the walls.
  case cmdLabyrinthChartresR:
    if (nMode == 1) {
      bT = NULL;
      if (!FBitmapCopy(bm.a, &bT))
        break;
      FFill(bm.a, 0, MY(bm.a)-1, fOn);
      BitmapXor(bm.a, bT);
      BitmapReverse(bm.a);
      DeallocateBitmap(bT);
      ws.wCmdSpree = cmdPolishMaze;
    }
    break;
  }

  if (nMode == 0) {
    SetXyh(bm.a);
  }
}


// Fire a macro event if defined, passing in the given parameters and
// returning the result of the macro's execution.

flag FCheckEvent(int iEvent, int x, int y, int z)
{
  if (iEvent <= 0 || !FEnsureLVar(cLetter))
    return fFalse;

  // Store parameters in custom variables @x, @y, and @z.
  ws.rglVar[iLetterX] = x; ws.rglVar[iLetterY] = y; ws.rglVar[iLetterZ] = z;
  RunMacro(iEvent);

  // Return the result that the macro should have stored in @z.
  return ws.rglVar[iLetterZ] != 0;
}


// Fire a macro event that checks whether the dot is allowed to move. If not,
// the dot is moved back to its old location.

void CheckEventMove(int iEvent, int xold, int yold, int zold, int eold,
  flag fUndo)
{
  int xnew = dr.x, ynew = dr.y, znew = dr.z, enew = dr.zElev;

  // Note if the position of the dot was changed by the event macro, that
  // position is left alone, regardless of the return value.
  dr.x = xold; dr.y = yold; dr.z = zold; dr.zElev = eold;
  if (FCheckEvent(iEvent, xnew, ynew, znew) != fUndo &&
    dr.x == xold && dr.y == yold && dr.z == zold && dr.zElev == eold) {
    dr.x = xnew; dr.y = ynew; dr.z = znew; dr.zElev = enew;
  }
}


// Fire a macro event that checks whether the dot is allowed to move to its
// current location and offset within that pixel, returning yes or no.

flag FCheckEventMove2(int iEvent, int xold, int yold, int zold,
  int xold2, int yold2, int zold2)
{
  int xnew = dr.x, ynew = dr.y, znew = dr.z,
    xnew2 = dr.nOffsetX, ynew2 = dr.nOffsetY, znew2 = dr.nOffsetZ;
  flag fRet;

  dr.x = xold; dr.y = yold; dr.z = zold;
  dr.nOffsetX = xold2; dr.nOffsetY = yold2; dr.nOffsetZ = zold2;
  fRet = FCheckEvent(iEvent, xnew, ynew, znew);
  dr.x = xnew; dr.y = ynew; dr.z = znew;
  dr.nOffsetX = xnew2; dr.nOffsetY = ynew2; dr.nOffsetZ = znew2;
  return fRet;
}


// Fire a macro event for a mouse button click. The return value in @z is
// whether to skip doing the normal action associated with that click.

flag FCheckEventClick(int iEvent, long lParam)
{
  int x, y;

  if (iEvent <= 0 || !FEnsureLVar(cLetter))
    return fFalse;
  x = WLo(lParam); y = WHi(lParam);
  if (!dr.fInside) {

    // For overhead view, @x and @y contain the bitmap pixel clicked on.
    x -= bm.xpOrigin; y -= bm.ypOrigin;
    ws.rglVar[iLetterX] = x / bm.xpPoint + bm.xaOrigin - (x < 0);
    ws.rglVar[iLetterY] = y / bm.ypPoint + bm.yaOrigin - (y < 0);
  } else {

    // For perspective inside view, parameters @x and @y are scaled
    // coordinates running from (-10000, -10000) at the upper left corner of
    // the window to (10000, 10000) at bottom right.
    x = NMultDiv(x, 20000, ws.xpClient);
    y = NMultDiv(y, 20000, ws.ypClient);
    ws.rglVar[iLetterX] = x - 10000; ws.rglVar[iLetterY] = y - 10000;
  }

  // The parameter in @z is the raw pixel coordinates within the window.
  ws.rglVar[iLetterZ] = lParam;
  RunMacro(iEvent);
  return ws.rglVar[iLetterZ] != 0;
}


// Handle a mouse click on the window in perspective inside view.

void MouseClickInside(int x, int y)
{
  int i;

  // Based on the quadrant of the window the mouse was clicked in, move the
  // dot forward or backward, or rotate in place.
  y = NMultDiv(y, ws.xpClient, ws.ypClient);
  i = x > y ? (ws.xpClient - x > y ? 1 : 5) :
    (x > ws.xpClient - y ? 7 : 3);
  if (dr.f3D) {

    // For 3D bitmaps, clicking in the top 1/4 of the top quadrant will move
    // up, and clicking in bottom 1/4 of the bottom quadrant will move down.
    if (i == 1) {
      if (y < ws.xpClient / 8)
        i = 9;
    } else if (i == 7) {
      if (y > ws.xpClient * 7 / 8)
        i = 10;
    }
  }
  DoCommand(rgcmdMouse[i]);
}


// Return whether there's a wall at a pixel. If so, return the height in
// inside units of the bottom and top of that wall.

flag FGetWallVar(int x, int y, int z, int *pzLo, int *pzHi)
{
  int zLo, zHi;
  bitmap bWallVar;
  colmap cWallVar;
  KV kv;

  // If no wall at this pixel, don't need to do anything more.
  if (!Mg3(bm.a, x, y, z, dr.f3D))
    return fFalse;

  if (dr.nWallVar >= 0 && ws.cbMask > dr.nWallVar && ws.ccTexture >
    dr.nWallVar && (bWallVar = ws.rgbMask[dr.nWallVar]) != NULL &&
    (cWallVar = ws.rgcTexture[dr.nWallVar]) != NULL &&
    Mg3(bWallVar, x, y, z, dr.f3D)) {

    // For variable height walls, translate from percentage to inside units.
    kv = Cg3F(cWallVar, x, y, z, dr.f3D);
    zLo = UdD(kv) * dr.zWall >> iVarScale;
    zHi = UdU(kv) * dr.zWall >> iVarScale;
    if (zLo > zHi)
      SwapN(zLo, zHi);
  } else {

    // Non-variable height walls are always the default wall height.
    zLo = 0;
    zHi = dr.zWall;
  }
  *pzLo = zLo; *pzHi = zHi;
  return fTrue;
}


// Adjust the dot to the middle of the pixel it's over. If wall climbing is
// active, set the viewing height so the dot is on top of whatever wall.

void CenterWithinCell()
{
  int zLo;

  dr.nOffsetX = dr.nOffsetY = dr.nOffsetZ = dr.nOffsetD = 0;
  if (dr.fWall && dr.zStep >= 0) {
    if (!FGetWallVar(dr.x, dr.y, dr.z, &zLo, &dr.zElev))
      dr.zElev = 0;
  }
}


// Return whether the dot is blocked from moving to the given coordinates. If
// not, and wall climbing is active, set zElev to what the new viewing height
// would be after moving there.

flag FHitWall(int x, int y, int z, int *pzElev)
{
  int zLo, zHi;

  // If dot can move through walls, never blocked by anything.
  if (!dr.fWall)
    return fFalse;

  // If no wall climbing, then always and only blocked by walls.
  if (dr.zStep < 0)
    return Mg3Fast(bm.a, x, y, z, dr.f3D);

  // Check the wall's height, and if dot can step on top of or under it.
  if (FGetWallVar(x, y, z, &zLo, &zHi)) {
    if ((dr.zElev + dr.zWall) >> 1 < zLo && dr.zElev < zHi - dr.zElev)
      zHi = 0;
  } else
    zHi = 0;
  if (zHi - dr.zElev > dr.zStep)
    return fTrue;
  if (dr.zStep2 >= 0 && dr.zElev - zHi > dr.zStep2)
    return fTrue;
  if (pzElev != NULL)
    *pzElev = zHi;
  return fFalse;
}


// The dot has moved in one of the free movement modes of the perspective
// inside view. Undo to the dot's previous positioning if necessary.

flag FUndoIfHitWall(int xold, int yold, int zold,
  int xinold, int yinold, int zinold)
{
  int xwall, ywall, zwall, zElev, zElevX, zElevY, cDialog;
  flag fCheck, fHitX, fHitY;

  // If dot can move through walls, don't need to check anything.
  if (!dr.fWall)
    goto LCheckEdge;

  // If dot has landed on a wall pixel it can't move on or under, undo.
  if (FHitWall(dr.x, dr.y, dr.z, &zElev)) {
    xwall = dr.x; ywall = dr.y; zwall = dr.z;
    goto LUndoWall;
  }

  // If dot has moved one orthogonal pixel, don't need to check any more.
  if (NAbs(dr.x - xold) + NAbs(dr.y - yold) + NAbs(dr.z - zold) <= 1)
    goto LCheckEdge;

  // If dot has moved one diagonal pixel, check if can hop corners.
  if (dr.z == zold && NAbs(dr.y - yold) == 1 && NAbs(dr.x - xold) == 1) {
    if (dr.fNoCorner &&
      FHitWall(dr.x, yold, zold, NULL) && FHitWall(xold, dr.y, zold, NULL)) {
      xwall = dr.x; ywall = yold; zwall = dr.z;
      goto LUndoWall;
    }
    goto LCheckEdge;
  }

  // Dot has moved at least two pixels. Check the middle pixels.
  xwall = (xold + dr.x) >> 1; ywall = (yold + dr.y) >> 1;
  zwall = (zold + dr.z) >> 1;
  if ((xwall != xold || ywall != yold || zwall != zold) &&
    FHitWall(xwall, ywall, zwall, NULL)) {
    if (dr.fNoCorner)
      goto LUndoWall;
  } else if (!dr.fNoCorner)
    goto LCheckEdge;
  xwall = (xold + dr.x + 1) >> 1; ywall = (yold + dr.y + 1) >> 1;
  zwall = (zold + dr.z + 1) >> 1;
  if ((xwall != xold || ywall != yold || zwall != zold) &&
    FHitWall(xwall, ywall, zwall, NULL))
    goto LUndoWall;

LCheckEdge:
  // Didn't hit a wall. Check if stepped off the bitmap.
  if (FLegalLevel(bm.a, xold, yold, dr.f3D) &&
    !FLegalLevel(bm.a, dr.x, dr.y, dr.f3D) &&
    (dr.nEdge == nEdgeStop || FCheckEventMove2(ws.iEventEdge,
    xold, yold, zold, xinold, yinold, zinold)))
    goto LUndo;
  if (ws.iEventMove > 0 &&
    !dr.fInSmooth && (dr.x != xold || dr.y != yold || dr.z != zold) &&
    FCheckEventMove2(ws.iEventMove, xold, yold, zold, xinold, yinold, zinold))
    goto LUndo;
  if (dr.zStep >= 0 && dr.fWall)
    dr.zElev = zElev;
  return fFalse;

LUndoWall:
  // Hit a wall. Before undoing, fire the move into wall event.
  cDialog = ws.cDialog;
  SwapN(dr.x, xwall); SwapN(dr.y, ywall); SwapN(dr.z, zwall);
  fCheck = FCheckEventMove2(ws.iEventWall,
    xold, yold, zold, xinold, yinold, zinold);
  dr.x = xwall; dr.y = ywall; dr.z = zwall;
  if (fCheck)
    return fFalse;

  // Check if able to slide along the wall, instead of just undoing.
  if (dr.fGlance && cDialog == ws.cDialog) {
    fHitX = FHitWall(dr.x, yold, dr.z, &zElevX) || (NAbs(dr.x - xold) > 1 &&
      FHitWall((dr.x + xold) >> 1, yold, dr.z, NULL));
    fHitY = FHitWall(xold, dr.y, dr.z, &zElevY) || (NAbs(dr.y - yold) > 1 &&
      FHitWall(xold, (dr.y + yold) >> 1, dr.z, NULL));
    if (!fHitX || !fHitY) {
      if (fHitX || (!fHitX && !fHitY &&
        NAbs(dr.nOffsetX - xinold) > NAbs(dr.nOffsetY - yinold))) {
        dr.x = xold; dr.nOffsetX = xinold; zElev = zElevY;
      } else {
        dr.y = yold; dr.nOffsetY = yinold; zElev = zElevX;
      }
      goto LCheckEdge;
    }
  }

LUndo:
  dr.x = xold; dr.y = yold; dr.z = zold;
  dr.nOffsetX = xinold; dr.nOffsetY = yinold; dr.nOffsetZ = zinold;
  return fTrue;
}


// The dot has moved or rotated one unit. Do smooth animation from the old
// positioning to the current positioning. This implements the intermediate
// frames seen in the Smooth Movement inside display mode.

void DotSmoothAnimate(int wCmd, int xold, int yold, int zold, int dold)
{
  int xnew, ynew, znew, dnew, i, iMax, dMax, dsav, zMax, zsav, wCmdT;
  flag fRedrawAfter;

  dr.fInSmooth = fTrue;
  if (dr.fNarrow) {
    xold |= 1; yold |= 1; zold &= ~1; dr.x |= 1; dr.y |= 1; dr.z &= ~1;
  }
  xnew = dr.x; ynew = dr.y; znew = dr.z; dnew = dr.dir;
  dr.x = xold; dr.y = yold; dr.z = zold; dr.dir = dold;

  // Animate a rotation if the direction has changed.
  if (dnew != dold) {
    if (((dold + 1) & DIRS1) == dnew)
      wCmdT = cmdMove4;
    else if (((dold - 1) & DIRS1) == dnew)
      wCmdT = cmdMove6;
    else {

      // Turning around 180 degrees, i.e. at a dead end. If following the
      // right hand wall, rotate to the left. If following left hand wall,
      // rotate right, otherwise pick one at random.
      if (wCmd == cmdMoveRight)
        wCmdT = cmdMove4;
      else if (wCmd == cmdMoveLeft)
        wCmdT = cmdMove6;
      else
        wCmdT = Rnd(0, 1) ? cmdMove4 : cmdMove6;
    }
    iMax = NAbs(dnew - dold);
    if (iMax >= 3)
      iMax = 1;
    dMax = iMax*90;
    iMax *= dr.nFrameD;
    dsav = dr.nSpeedD;
    for (i = 0; i < iMax; i++) {
      dr.nSpeedD = (i+1)*dMax/iMax - i*dMax/iMax;
      DoCommand(wCmdT);
    }
    dr.nSpeedD = dsav;
  }

  // Animate movement if the location has changed.
  if (xnew != xold || ynew != yold) {
    if (wCmd != cmdJump4 && wCmd != cmdJump6) {
      if ((xold - xoff[dnew] == xnew && yold - yoff[dnew] == ynew) ||
        ((xold - xoff2[dnew] == xnew && yold - yoff2[dnew] == ynew)))
        wCmdT = cmdMove2;
      else
        wCmdT = cmdMove8;
    } else
      wCmdT = wCmd;
    zMax = (dr.zCell + (dr.fNarrow ? dr.zCellNarrow : dr.zCell)) *
      (NAbs(xnew - xold) + NAbs(ynew - yold)) / 2;
    iMax = dr.nFrameXY;
    zsav = dr.nSpeedXY;
    fRedrawAfter = dr.fRedrawAfter && ws.iEventMove2 > 0;
    for (i = 0; i < iMax - fRedrawAfter; i++) {
      dr.nSpeedXY = (i+1)*zMax/iMax - i*zMax/iMax;
      DoCommand(wCmdT);
    }
    if (fRedrawAfter) {
      dr.x = xnew; dr.y = ynew;
      dr.nOffsetX = dr.nOffsetY = 0;
    }
    dr.nSpeedXY = zsav;
  }

  // Animate a level change if the level height has changed.
  if (znew != zold) {
    wCmdT = zold - znew > 0 ? cmdMoveU : cmdMoveD;
    zMax = dr.zCell;
    iMax = dr.nFrameZ;
    zsav = dr.nSpeedZ;
    for (i = 0; i < iMax; i++) {
      dr.nSpeedZ = (i+1)*zMax/iMax - i*zMax/iMax;
      DoCommand(wCmdT);
    }
    dr.nSpeedZ = zsav;
  }
  dr.fInSmooth = fFalse;

  // An event may have stopped animation part way between cells.
  if (dr.nOffsetX != 0 || dr.nOffsetY != 0)
    SystemHook(hosRedraw);
}


// Given a window pixel (that the mouse is over) move the dot one unit toward
// that pixel, so the dot moves toward the mouse.

void DotMouseChase(int xp, int yp)
{
  int rgxi[4], rgyi[4], xi, yi, x, y, c, i, j;

  // Determine preferred offset to apply to dot
  x = xp - bm.xpDot;
  x = x / bm.xpPoint - (x < 0);
  y = yp - bm.ypDot;
  y = y / bm.ypPoint - (y < 0);
  xi = NSgn(x); yi = NSgn(y);
  if (xi == 0 && yi == 0)
    return;

  // Determine a few alternate offsets in case the first doesn't work
  c = 3 + (xi != 0 && yi != 0);
  for (i = 0; i < c; i++) {
    rgxi[i] = xi; rgyi[i] = yi;
  }
  if (c >= 4) {
    if (NAbs(x) >= NAbs(y)) {
      rgyi[1] = 0; rgxi[2] = 0; neg(rgyi[3]);
    } else {
      rgxi[1] = 0; rgyi[2] = 0; neg(rgxi[3]);
    }
    if (NAbs(x) == NAbs(y)) // Pure diagonal case only has three options
      c = 3;
  } else {
    if (xi == 0) {
      rgxi[1] = -1; rgxi[2] = 1;
    } else {
      rgyi[1] = -1; rgyi[2] = 1;
    }
    if (NAbs(x + y) <= 1) // Don't vibrate back and forth next to on pixel
      c = 1;
  }

  // Test each offset until one free and can move in that direction
  for (i = 0; i < c; i++) {
    xi = rgxi[i]; yi = rgyi[i];
    if (dr.fNoDiagonal && xi != 0 && yi != 0)
      continue;
    if (dr.fGlance && !dr.fInside) {
      if (FHitWall(dr.x + xi, dr.y + yi, dr.z, NULL))
        continue;
      if (dr.fNoCorner && xi != 0 && yi != 0 &&
        FHitWall(dr.x + xi, dr.y, dr.z, NULL) &&
        FHitWall(dr.x, dr.y + yi, dr.z, NULL) && !dr.fTwo)
        continue;
    }
    j = (yi + 1) * 3 + (xi + 1);
    Assert(j != 4);
    DoCommand(rgcmdMouse[j]);
    break;
  }
}


// Given a range of characters containing a list of space separated strings,
// add the strings to a lookup trie. Implements the DefineConst operation.

flag FCreateConstTrie(CONST char *szIn, int cchIn)
{
  char szStrs[cchSzOpr*2], **rgsz = NULL, *pch;
  int csz = 0, csMax = 16000, isz, cs, i, n;
  flag fRet = fFalse, fPlus;

  // Count strings
  CopyRgchToSz(szIn, cchIn, szStrs, cchSzOpr*2);
  for (pch = szStrs; *pch; pch += (*pch != chNull)) {
    if (!FDigitCh(*pch) && *pch != '+')
      csz++;
    while (*pch && *pch != ' ')
      pch++;
  }

  // Compose list of strings
  rgsz = RgAllocate(csz, char *);
  if (rgsz == NULL)
    goto LDone;
  if (ws.rgnTrieConst != NULL)
    DeallocateP(ws.rgnTrieConst);
  ws.rgnTrieConst = RgAllocate(csz, int);
  if (ws.rgnTrieConst == NULL)
    goto LDone;
  pch = szStrs;
  for (isz = 0, i = 0; isz < csz; isz++, i++) {

    // Special: A number redefines lookup result for the following strings.
    // Special: +number adds delta to current lookup result for next strings.
    while (FDigitCh(*pch) || *pch == '+') {
      fPlus = (*pch == '+');
      pch += fPlus;
      for (n = 0; *pch && *pch != ' '; pch++)
        n = n * 10 + (*pch - '0');
      i = fPlus ? i + n - 1 : n;
      if (*pch)
        pch++;
    }
    rgsz[isz] = pch;
    ws.rgnTrieConst[isz] = i;
    while (*pch && *pch != ' ')
      pch++;
    if (*pch)
      *pch++ = chNull;
  }

  // Compose trie
  if (ws.rgsTrieConst != NULL)
    DeallocateP(ws.rgsTrieConst);
  ws.rgsTrieConst = RgAllocate(csMax, short);
  if (ws.rgsTrieConst == NULL)
    goto LDone;
  cs = CsCreateTrie((CONST uchar **)rgsz, csz, ws.rgsTrieConst, csMax,
    fFalse);

  // Sanity check trie to ensure all string lookups return expected results
  for (isz = 0; isz < csz; isz++) {
    i = ILookupTrieSz(ws.rgsTrieConst, rgsz[isz], fFalse);
    if (i != isz) {
      PrintSzNN_E("Input string %d maps to index %d.", isz, i);
      goto LDone;
    }
  }

  fRet = fTrue;
LDone:
  if (rgsz != NULL)
    DeallocateP(rgsz);
  return fRet;
}


#define nTrib 24      // Number of tributes
#define nInv 10       // Inventory size
#define zVis 16       // Visibility distance
#define zRng 8        // Throwable weapon range distance
#define dobj 13       // Ammo to launcher offset
#define nFoodMax 500  // Maximum food allowed
#define nHealthMax 10 // Maximum health allowed
#define zClimb 128    // Vertical step units

enum {
  bmpVar  = 1, // Wall height bitmap
  bmpInv  = 4, // Inventory bitmap
  bmpAlly = 5, // Faction bitmap
  bmpMap  = 6, // Arena copy bitmap
  bmpH20  = 7, // Water location bitmap
};

enum {
  numTrib  = 19, // Tribues left
  numTube  = 43, // Launch tube color
  numCorn  = 44, // Cornucopia color
  numMaze  = 45, // Maze color
  numCoor  = 100, // Tribute X & Y coordinates
  numElev  = 150, // Tribute Z coordinate
  numHP    = 200, // Tribute hitpoints
  numFood  = 250, // Tribute food
  numKill  = 300, // Tribute kills
  numWield = 350, // Tribute wielded weapon index
  numNet   = 400, // Tribute netting
  numDest  = 450, // Tribute destination coordinates
  numDmg   = 500, // Weapon damage
  numUsage = 525, // Weapon usage
  numDist  = 550, // Weapon favored district
  numX     = 21, // X increments
  numY     = 22, // Y increments
};

enum {
  actWait  = 0, // Do nothing and wait
  actUse   = 1, // Use item
  actGet   = 2, // Pick up whatever standing over
  actMove  = 3, // Move or melee attack in direction
  actThrow = 4, // Throw item in direction
  actDrop  = 5, // Drop item
  actDown  = 6, // Climb down
  actUp    = 7, // Climb up
  actGive  = 8, // Give item in direction
  actDig   = 9, // Dig
};

enum {
  objFist  = 0,  // Fist weapon
  objNet   = 4,  // Net weapon
  objClub  = 5,  // Club weapon
  objPick  = 7,  // Pick weapon
  objSpear = 10, // Spear weapon
  objAxe   = 11, // Axe weapon
  objTrid  = 13, // Trident weapon
  objBerry = 17, // Berries
  objNight = 18, // Nightlock berries
  objBread = 19, // Loaf of bread
  objFish  = 20, // Fish
  objHeal  = 21, // First-aid kit
  objMine0 = 22, // Landmine (inactive)
  objMine1 = 23, // Landmine (active)
  objTeeth = 24, // Mutt teeth
  objMax   = 25, // Number of objects
};

enum {
  desPunch   = 0,
  desUp      = 1,
  desDown    = 9999,
  desStart   = 10000,
  desChase   = 10000,
  desGive    = 30000 + 100,
  desDig     = 30000 + 110,
  desMine    = 30000 + 120,
  desDrop    = 30000 + 130,
  desEat     = 30000 + 140,
  desHeal    = 30000 + 150,
  desWield   = 30000 + 160,
  desGet     = 30000 + 170,
  desMelee   = 30000 + 500,
  desThrow   = 30000 + 500,
  desFlee    = 40000,
  desGiveHi  = 60000 + 950,
  desGetHi   = 60000 + 960,
  desWieldHi = 60000 + 970,
  desHealHi  = 60000 + 980,
  desEatHi   = 60000 + 990,
};

#define Xoff(d) (ws.rgszVar[numX][d] - '1')
#define Yoff(d) (ws.rgszVar[numY][d] - '1')
#define ZElev(n) ws.rglVar[numElev + (n)]
#define NWater(x, y) \
  (Mg(ws.rgbMask[bmpH20], x, y) * RgbB(Cg(ws.rgcTexture[bmpMap], x, y)))
#define NDist(i) (((i) >> 1) + 1)
#define NWield(i) ws.rglVar[numWield + (i)]
#define FObjInInv(o) ((grfObjInv & (1 << (o))) != 0)
#define ObjInvN(i, n) ((i) < 0 ? 0 : Cg(ws.rgcTexture[bmpInv], n, i))
#define ObjInv(i) ObjInvN(i, nSelf)
#define WepDmgN(o, n) \
  (ws.rglVar[numDmg + (o)] + (ws.rglVar[numDist + (o)] == NDist(n))*2)
#define WepDmg(o) WepDmgN(o, nSelf)
#define FactionN(i, n) Cg(ws.rgcTexture[bmpAlly], n, i)
#define Faction(i) FactionN(i, nSelf)
#define FAlly(n) ((n) <= 100)
#define FWep(o)       ((o) != objFist && ws.rglVar[numUsage + (o)] > 0)
#define FWepMelee(o)  ((o) != objFist && (ws.rglVar[numUsage + (o)] & 9) == 1)
#define FWepThrow(o)  ((o) != objFist && (ws.rglVar[numUsage + (o)] & 2) > 0)
#define FWepAmmo(o)   ((o) != objFist && (ws.rglVar[numUsage + (o)] & 4) > 0)
#define FWepLaunch(o) ((o) != objFist && (ws.rglVar[numUsage + (o)] & 8) > 0)
#define FConsum(o) \
  ((o) == objBerry || (o) == objBread || (o) == objHeal || (o) == objFish)
#define FFisher(o) ((o) == objNet || (o) == objSpear || (o) == objTrid)
#define FStronger(n) (ws.rglVar[numNet + (n)] > 0 || \
  (nHealth * 2 / WepDmgN(ObjInvN(NWield(n), n), n)) > \
  (ws.rglVar[numHP + (n)] * 2 / nDmgWield))
#define Desire(a, i, d, p) if ((p) > nPriority) \
  { nAction = (a); nItem = (i); nDir = (d); nPriority = (p); }

// Determine the best move for a computer controlled Hunger Games tribute.
// Contains AI specialized to the Hunger Games script, and implements the
// Hunger operation. Loop over all possible actions, assigning a value to
// each, then select whichever potential action has the highest value.

flag FHungerGame(int nSelf)
{
  bitmap b = NULL;
  int nAction = actWait, nItem = 0, nDir = 0, nPriority = -1,
    rgnMove[DIRS2], rgnDesObj[objMax], rgnConObj[objMax], rgnValObj[objMax],
    rgnDesDir[DIRS2], cTribute, nDistrict, xSelf, ySelf, zSelf, xCorn, yCorn,
    grfObjInv = 0, iWield, objWield, objHere, nHealth, nFood, iItem, nTime,
    nMove, nDes, nDmg, nDmgMax, nDmgWield, iItemMax, cInvFree = 0,
    cConsum = 0, dGive = -1, nGive = 255, obj, x, y, d, x2, y2, d2, i, j, l;
  flag fMutt, fNet, fWater, fCanClimb, fCanFish, fCanSwim,
    fMelee = fFalse, fRanged = fFalse, fSafe;

  // Ensure valid game space
  if (bm.c == NULL || bm.c2 == NULL || bm.c3 == NULL ||
    ws.rgcTexture == NULL || ws.ccTexture <= bmpMap ||
    ws.rgcTexture[bmpInv] == NULL || ws.rgcTexture[bmpAlly] == NULL)
    return fFalse;
  if (!FEnsureLVar(numDist) || !FEnsureSzVar(numY) ||
    ws.rgszVar[numX] == NULL || ws.rgszVar[numY] == NULL ||
    CchSz(ws.rgszVar[numX]) != 8 || CchSz(ws.rgszVar[numY]) != 8)
    return fFalse;
  i = (zVis << 1) + 1;
  if (!FAllocateBitmap(&b, i, i, NULL))
    return fFalse;

  // Setup initial variables
  cTribute = ws.rglVar[numTrib];
  nDistrict = NDist(nSelf);
  fMutt = nDistrict > 12;
  l = ws.rglVar[numCoor + nSelf];
  xSelf = UdU(l); ySelf = UdD(l);
  zSelf = ZElev(nSelf);
  nTime = ws.rglVar[16];
  nHealth = ws.rglVar[numHP + nSelf];
  nFood = ws.rglVar[numFood + nSelf];
  fNet = ws.rglVar[numNet + nSelf] > 0;
  fWater = NWater(xSelf, ySelf) > 0;
  iWield = NWield(nSelf);
  objWield = ObjInv(iWield);
  nDmgWield = WepDmg(objWield);
  fCanFish = nDistrict == 4 || FFisher(objWield);
  fCanSwim = nDistrict == 4 || objWield == objClub;
  objHere = RgbG(Cg(bm.c3, xSelf, ySelf)) % 100;
  for (iItem = 0; iItem < nInv; iItem++) {
    obj = ObjInv(iItem);
    grfObjInv |= (1 << obj);
    cInvFree += (obj == 0);
    cConsum += FConsum(obj);
  }
  if (FWepLaunch(objWield) && (!FObjInInv(objWield-dobj) || fNet))
    nDmgWield = 1;
  l = ws.rglVar[numDest + nSelf];
  xCorn = UdU(l); yCorn = UdD(l);

  // Setup what's next to us
  for (d = 0; d < DIRS; d++) {
    x = xSelf + xoff[d];
    y = ySelf + yoff[d];
    if (Mg(bm.a, x, y) && !Mg(bm.a3, x, y) &&
      Cg(ws.rgcTexture[bmpVar], x, y) >= zSelf + zClimb*3/2)
      break;
  }
  fCanClimb = d < DIRS && !fMutt;
  for (d = 0; d < DIRS2; d++) {
    rgnMove[d] = -1;
    x = xSelf + Xoff(d);
    y = ySelf + Yoff(d);
    if (!FLegal(bm.c, x, y))
      continue;
    l = Cg(bm.c3, x, y);
    obj = RgbG(l) % 100;
    l = RgbR(l);
    // Never step onto activated landmines
    if (obj == objMine1 && zSelf <= 0 && !fMutt)
      continue;
    // Never step into solid objects (unless can chop or dig it)
    if (Mg(bm.a, x, y) && l == 0) {
      j = Cg(bm.c, x, y);
      if (j == ws.rglVar[numCorn] || j == ws.rglVar[numMaze] ||
        !((nDistrict == 7 && objWield == objAxe && RgbG(j) > RgbR(j)) ||
        (nDistrict == 2 && objWield == objPick && RgbG(j) == RgbB(j))))
        continue;
    }
    // Never step into deep water if can't swim
    i = NWater(x, y);
    if (i > 0 && i <= 170 && !fCanSwim && l == 0)
      continue;
    // Never fall out of a tree
    if (zSelf > zClimb) {
      for (d2 = 0; d2 < DIRS; d2++) {
        x2 = x + xoff[d2];
        y2 = y + yoff[d2];
        if (Mg(bm.a, x2, y2) && !Mg(bm.a3, x2, y2) &&
          Cg(ws.rgcTexture[bmpVar], x2, y2) >= zSelf + zClimb)
          break;
      }
      if (d2 >= DIRS)
        continue;
    }
    if (l == 0)
      rgnMove[d] = 0;
    else {
      l %= 100;
      i = Faction(l);
      rgnMove[d] = i;
      if (!FAlly(i))
        fMelee = fTrue;
    }
  }

  // Setup how interested in objects
  for (obj = 0; obj < objMax; obj++) {
    nDes = 0;
    if (FWep(obj)) {
      // Don't want non-throwable ammo unless already have launcher
      // Don't want melee only weapon worse than melee weapon wielded
      nDmg = WepDmg(obj);
      if (FWepLaunch(obj))
        nDes = FObjInInv(obj-dobj) ? nDmg : 1;
      else if (FWepAmmo(obj))
        nDes = FObjInInv(obj+dobj) ? WepDmg(obj+dobj) : nDmg * FWepThrow(obj);
      else if (FWepThrow(obj) || WepDmg(obj) >= nDmgWield)
        nDes = nDmg;
    } else {
      // Never want Nightlock
      if (obj == objHeal)
        nDes = nHealth < nHealthMax ? 9 : 5;
      else if (obj == objBread)
        nDes = nFood < nFoodMax ? 8 : 4;
      else if (obj == objBerry)
        nDes = nFood < nFoodMax ? 6 : 2;
      else if (obj == objMine0)
        nDes = nDistrict == 3 ? 5 : 1;
      else if (obj == objFish)
        nDes = nFood < nFoodMax ? 7 : 3;
    }
    if (fMutt)
      nDes = 0;
    rgnValObj[obj] = nDes;
    // No duplicate launcher or duplicate melee only weapon
    if ((FWepLaunch(obj) || (FWepMelee(obj) && !FWepThrow(obj))) &&
      FObjInInv(obj))
      nDes = 0;
    // No melee only weapon equal to melee only weapon already wielded
    if (FWepMelee(obj) && !FWepThrow(obj) && FWepMelee(objWield) &&
      !FWepThrow(objWield) && WepDmg(obj) <= nDmgWield)
      nDes = 0;
    rgnConObj[obj] = nDes;
    /* Save last inventory slot for consumables, and don't want anything at
      all if inventory full (unless need First-aid). */
    if ((cInvFree <= 0 && !(obj == objHeal && nHealth < nHealthMax)) ||
      (cInvFree <= 1 && cConsum <= 0 && !FConsum(obj)))
      nDes = 0;
    rgnDesObj[obj] = nDes;
  }

  // Look around and setup how interested in directions
  BitmapOff(b);
  for (y = ySelf-zVis; y <= ySelf+zVis; y++)
    for (x = xSelf-zVis; x <= xSelf+zVis; x++)
      if (Mg(bm.a, x, y) && !Mg(bm.a3, x, y))
        MP(b, x-xSelf+zVis, y-ySelf+zVis, fOn);
      else {
        i = NWater(x, y);
        if (i > 0 && i <= 170 && !fCanSwim)
          MP(b, x-xSelf+zVis, y-ySelf+zVis, fOn);
      }
  BitmapSeen(b, zVis, zVis, zVis);
  ClearPb(rgnDesDir, DIRS2 * sizeof(int));
  for (y = ySelf-zVis; y <= ySelf+zVis; y++)
    for (x = xSelf-zVis; x <= xSelf+zVis; x++) {
      if (MG(b, x-xSelf+zVis, y-ySelf+zVis))
        continue;
      nDes = 0;
      l = Cg(bm.c3, x, y);
      obj = RgbG(l) % 100;
      // Ignore objects in deep water if can't swim
      i = NWater(x, y);
      if (i > 0 && i <= 170 && !fCanSwim)
        obj = 0;
      // Ignore fish if can't catch them
      if (obj == objFish && i > 0 && !fCanFish)
        obj = 0;
      if (obj > 0) {
        i = rgnDesObj[obj];
        if (i > 0)
          nDes = i + (zVis - NMax(NAbs(x - xSelf), NAbs(y - ySelf))) * 10;
      } else {
        l = RgbR(l);
        if (l != 0) {
          l %= 100;
          if (fMutt)
            nDes = (l < nTrib)*3; // Mutts seek out all non-Mutts
          else if (FAlly(Faction(l)))
            nDes = 1; // Seek out allies
          else
            nDes = FStronger(l)*3; // Seek out weaker enemies
        }
      }
      if (nDes > 0)
        for (d = 0; d < DIRS2; d++) {
          x2 = xSelf + Xoff(d);
          y2 = ySelf + Yoff(d);
          if (NMax(NAbs(x - x2), NAbs(y - y2)) <
            NMax(NAbs(x - xSelf), NAbs(y - ySelf)))
            rgnDesDir[d] = NMax(rgnDesDir[d], nDes);
        }
    }

  // Seek out coordinates
  if (nTime > 25) {
    // Don't if follower in alliance, i.e. if earlier allied tribute nearby
    for (i = 0; i < nSelf; i++) {
      if (ws.rglVar[numHP + i] > 0 && FAlly(Faction(i))) {
        l = ws.rglVar[numCoor + i];
        x = UdU(l); y = UdD(l);
        if (NMax(NAbs(xSelf - x), NAbs(ySelf - y)) <= zRng)
          break;
      }
    }
    if (i >= nSelf || fMutt) {
      i = 0;
      for (d = 0; d < DIRS2; d++) {
        if (rgnMove[d] < 0)
          continue;
        x = xSelf + Xoff(d);
        y = ySelf + Yoff(d);
        if (NMax(NAbs(xCorn - x), NAbs(yCorn - y)) <
          NMax(NAbs(xCorn - xSelf), NAbs(yCorn - ySelf))) {
          rgnDesDir[d] += Rnd(1, 2);
          i++;
        }
      }
      // If can't make progress, indicate want new destination
      if (i < 1)
        ws.rglVar[numDest + nSelf] = UD(xSelf, ySelf);
    }
  }

  // Move or melee attack
  for (d = 0; d < DIRS2; d++) {
    nMove = rgnMove[d];
    if (nMove < 0)
      continue;
    x = xSelf + Xoff(d);
    y = ySelf + Yoff(d);
    // Consider move to an adjacent space
    if (nMove == 0) {
      nDes = rgnDesDir[d]*10 + Rnd(1, 9);
      // Approach or flee cornucopia at start of game
      if (nTime < 25 &&
        NMax(NAbs(xCorn - x), NAbs(yCorn - y)) <
        NMax(NAbs(xCorn - xSelf), NAbs(yCorn - ySelf)))
        nDes += desStart;
      // Flee stronger non-allies if unarmed
      if (fMelee && nDmgWield <= 1 && !fMutt) {
        i = desFlee;
        for (d2 = 0; d2 < DIRS2; d2++) {
          x2 = x + Xoff(d2);
          y2 = y + Yoff(d2);
          l = RgbR(Cg(bm.c3, x2, y2));
          if (l != 0) {
            l %= 100;
            if (!FAlly(Faction(l)) && !FStronger(l))
              i = 0;
          }
        }
        nDes += i;
      // Chase weaker non-allies if armed
      } else if (!fMelee && nDmgWield > 1 && objWield > 0) {
        i = 0;
        for (d2 = 0; d2 < DIRS2; d2++) {
          x2 = x + Xoff(d2);
          y2 = y + Yoff(d2);
          l = RgbR(Cg(bm.c3, x2, y2));
          if (l != 0) {
            l %= 100;
            if (!FAlly(Faction(l)) && (fMutt || FStronger(l)))
              i = desChase;
          }
        }
        nDes += i;
      }
    // Consider attacking an adjacent tribute
    } else {
      l = RgbR(Cg(bm.c3, x, y));
      // Consider giving item to allied or neutral tribute
      if (l != 0) {
        j = l % 100;
        if (nMove < nGive) {
          for (i = iItem = 0; iItem < nInv; iItem++) {
            obj = ObjInvN(iItem, j);
            i += (obj == 0);
          }
          if (i >= 2 &&
            (FAlly(nMove) || (cTribute > 2 && FAlly(nMove - 50)))) {
            nGive = nMove;
            dGive = d;
          }
        }
      }
      // Don't ever attack allies
      if (FAlly(nMove))
        continue;
      // Can't attack more than one square above or below
      if (l != 0) {
        j = l % 100;
        if (NAbs(zSelf - ZElev(j)) > zClimb) {
          if (FStronger(j)) {
            if (zSelf > ZElev(j)) {
              Desire(actDown, 0, 0, desChase + 1);
            } else if (fCanClimb)
              Desire(actUp, 0, 0, desChase);
          }
          continue;
        }
      }
      nDes = desMelee + nMove;
      // Don't attack with fist unless cornered
      if (nDmgWield <= 1 || FWepLaunch(objWield))
        nDes = desPunch;
    }
    Desire(actMove, 0, d, nDes);
  }

  // Throw attack
  iItemMax = nDmgMax = -1;
  for (iItem = nInv-1; iItem >= 0; iItem--) {
    // Look for ammo to launch with wielded launcher (1st priority)
    obj = ObjInv(iItem);
    nDmg = FWepAmmo(obj) && objWield == obj+dobj ? WepDmg(obj+dobj) : -1;
    if (nDmg > nDmgMax) {
      iItemMax = iItem;
      nDmgMax = nDmg;
    }
  }
  for (iItem = nInv-1; iItem >= 0; iItem--) {
    // Look for unwielded throwable weapon (2nd priority)
    obj = ObjInv(iItem);
    nDmg = FWepThrow(obj) && (!FWepMelee(obj) || iItem != iWield) ?
      WepDmg(obj) : -1;
    if (nDmg > nDmgMax) {
      iItemMax = iItem;
      nDmgMax = nDmg;
    }
  }
  for (d = 0; d < DIRS2; d++) {
    // If best item thrown in each direction, what would be hit?
    x = xSelf; y = ySelf;
    for (j = 0; j < zRng; j++) {
      x += Xoff(d);
      y += Yoff(d);
      if (!FLegal(bm.c, x, y))
        break;
      l = RgbR(Cg(bm.c3, x, y));
      if (Mg(bm.a, x, y) && l == 0)
        break;
      if (l == 0)
        continue;
      l %= 100;
      i = Faction(l);
      if (FAlly(i)) // Don't ever attack allies
        break;
      fRanged = fTrue;
      if (iItemMax >= 0 && !fNet)
        Desire(actThrow, iItemMax, d, desThrow + i);
    }
  }
  fSafe = !fMelee && !fRanged;

  // Climb
  if (zSelf > 0 && fSafe && (zSelf > zClimb || objHere != objMine1))
    Desire(actDown, 0, 0, desDown);
  if (fCanClimb)
    Desire(actUp, 0, 0, desUp);

  // Dig
  if (zSelf <= 0 && !fWater && objHere == 0 && objWield == objPick && fSafe &&
    cInvFree > 1) {
    for (d = 0; d < DIRS2; d++)
      if (Cg(ws.rgcTexture[bmpMap], xSelf + xoff[d], ySelf + yoff[d]) ==
        ws.rglVar[numTube])
        break;
    if (nDistrict == 12 || d < DIRS2) {
      l = Cg(ws.rgcTexture[bmpMap], xSelf, ySelf);
      if (l != ws.rglVar[numTube] && (l | 0x000f1f) != kvBrown)
        Desire(actDig, 0, 0, desDig);
    }
  }

  // Pick up item
  if (cInvFree > 0 && zSelf <= 0 && !fNet && rgnDesObj[objHere] > 0 &&
    !(objHere == objFish && fWater && !fCanFish))
    Desire(actGet, 0, 0, (FWep(objHere) && rgnDesObj[objHere] >= 5) ||
      (FWepMelee(objHere) && nDmgWield <= 1) ? desGetHi : desGet);

  // Wield weapon
  for (iItem = 0; iItem < nInv; iItem++) {
    obj = ObjInv(iItem);
    if (!FWep(obj) || iItem == iWield)
      continue;
    nDmg = WepDmg(obj);
    if (FWepLaunch(obj) && (!FObjInInv(obj-dobj) || fNet))
      nDmg = 1;
    if (FWepLaunch(obj)) {
      /* Wield launcher if have ammo for it and better than what currently
        have, or if equal to wielded melee weapon and not in melee. */
      if (FObjInInv(obj-dobj) && (nDmg > nDmgWield ||
        (nDmg >= nDmgWield && FWepMelee(objWield) && !fMelee)))
        Desire(actUse, iItem, 0, (fSafe ? desWield : desWieldHi) + nDmg);
      // Wield launcher if not wielding anything and not threatened.
      if (objWield < 1 && fSafe)
        Desire(actUse, iItem, 0, desWield + nDmg);
    }
    if (FWepMelee(obj)) {
      // Wield if strictly better than what currently have.
      if (nDmg > nDmgWield)
        Desire(actUse, iItem, 0, (fSafe ? desWield : desWieldHi) + nDmg);
      /* Wield if equal to what currently have, and can switch from throwable
        to melee only, e.g. Trident to Sword. */
      if (nDmg == nDmgWield &&
        !FWepThrow(obj) && FWepThrow(objWield) && fSafe)
        Desire(actUse, iItem, 0, desWield + nDmg);
    }
    // Wield lesser weapon if it can be used to catch fish here
    if (FFisher(obj) && objHere == objFish && !fCanFish && nFood < nFoodMax &&
      fSafe) {
      Desire(actUse, iItem, 0, desWield + nDmg);
    }
  }

  // Use, drop, or give item
  for (iItem = nInv-1; iItem >= 0; iItem--) {
    obj = ObjInv(iItem);
    // Heal self
    if (obj == objHeal && nHealth < nHealthMax) {
      if (nFood <= 1 && nHealth <= 1)
        Desire(actUse, iItem, 0, desHealHi);
      if (fSafe)
        Desire(actUse, iItem, 0, desHeal);
    }
    // Eat food
    if (obj == objBerry || obj == objBread || obj == objFish) {
      if (nFood <= 1 && nHealth <= 1)
        Desire(actUse, iItem, 0, desEatHi + (obj == objBread));
      if (nFood < nFoodMax && nTime > 40 + nSelf && fSafe)
        Desire(actUse, iItem, 0, desEat + (obj == objBread));
    }
    if (obj == objNight && nFood < nFoodMax && cTribute == 2) {
      for (i = 0; i < nTrib; i++) {
        if (i != nSelf && ws.rglVar[numHP + i] > 0 && Faction(i) <= 0)
          break;
      }
      if (i < nTrib)
        Desire(actUse, iItem, 0, desEatHi + 2);
    }
    // Place landmine
    if (fNet)
      continue;
    if (obj == objMine0 && objHere == 0 && fSafe)
      Desire(actUse, iItem, 0, desMine);

    // Drop item
    if (iItem == iWield || nTime <= 50 || (!fSafe && dGive < 0) ||
      (fMutt && obj == objTeeth))
      continue;
    // Always drop Nightlock as soon as possible
    if (obj == objNight)
      Desire(actDrop, iItem, 0, desDrop);
    /* Drop melee only weapon worse than melee weapon being wielded, or
      equal to melee only weapon already wielded. */
    if (FWepMelee(obj) && !FWepThrow(obj) && FWepMelee(objWield) &&
      objWield != objNet && (WepDmg(obj) < nDmgWield ||
      (WepDmg(obj) == nDmgWield && !FWepThrow(objWield)))) {
      if (dGive < 0 || !FAlly(nGive - ws.rglVar[numDmg + obj]*10) ||
        (FAlly(nGive) && Rnd(0, 1))) {
        Desire(actDrop, iItem, 0, desDrop);
      } else
        Desire(actGive, iItem, dGive, FAlly(nGive) ? desGive : desGiveHi);
    }
    // Improve full inventory if on top of better weapon or better food
    if (cInvFree <= 1) {
      // Don't drop launcher to pick up unthrowable ammo for that launcher
      // Don't drop anything for a fish if can't catch fish
      // Don't drop last consumable item in inventory
      if (obj > 0 && objHere > 0 && rgnValObj[obj] < rgnConObj[objHere] &&
        !(FWepAmmo(objHere) && !FWepThrow(objHere) && obj == objHere+dobj) &&
        !(objHere == objFish && fWater && !fCanFish) &&
        !(cConsum <= 1 && FConsum(obj))) {
        if (dGive < 0 || !FAlly(nGive - ws.rglVar[numDmg + obj]*10)) {
          Desire(actDrop, iItem, 0, desDrop - rgnValObj[obj]);
        } else
          Desire(actGive, iItem, dGive,
            (FAlly(nGive) ? desGive : desGiveHi) - rgnValObj[obj]);
      }
    }

    // Give item
    if (dGive >= 0 && !FAlly(nGive) &&
      FAlly(nGive - ws.rglVar[numDmg + obj]*10))
      Desire(actGive, iItem, dGive, desGiveHi - rgnValObj[obj]);
  }

  // Pick best move of all possible
  ws.rglVar[iLetterX] = nItem;
  ws.rglVar[iLetterY] = nDir;
  ws.rglVar[iLetterZ] = nAction;
  DeallocateBitmap(b);
  return fTrue;
}

/* daedalus.cpp */
