/*
** Daedalus (Version 2.5) File: draw.cpp
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains advanced non-perspective graphics routines, unrelated to
** Mazes.
**
** Created: 6/15/1990.
** Last code change: 10/10/2013.
*/

#include <stdio.h>
#include <stdlib.h>
#include "util.h"
#include "graphics.h"
#include "color.h"
#include "threed.h"
#include "draw.h"


DS ds = {10, 10, 10, 10, 10, 0, 0, 0, fFalse,
  0, 0, 0, 0, 10, 10, fTrue, fTrue,
  fTrue, fFalse, fFalse, fTrue, fFalse, fFalse,
  1, 1, 200.0, 20.0, 20.0, 20.0,
  kvBlack, kvBlue, kvYellow, kvGray, kvGreen, kvRed,
  {100.0, 75.0, 50.0},
  0, 0, fFalse, xStart, 0.0, 0L};


/*
******************************************************************************
** Render Overview
******************************************************************************
*/

// Draw some or all of the faces of a solid orthographic block, with one face
// oriented toward you. Used by various FillOverview functions.

void FillCube(bitmap b, int x, int y, int xs, int ys, int zs, int dir,
  flag fX, flag fY, flag fZ)
{
  int i;

  // Draw the outline of the cube.
  if (fX) {
    LineY(b, x + ys*dir, y - ys, y - ys + zs, fOff);
    LineZ(b, x + dir, y + zs - 1, y - ys + zs + 1, dir, fOff);
  }
  if (fY) {
    LineX(b, x - (xs - 1)*dir, x - dir, y + zs, fOff);
    LineY(b, x - xs*dir, y, y + zs, fOff);
  }
  if (fZ) {
    LineX(b, x - (xs - ys)*dir, x + (ys - 1)*dir, y - ys, fOff);
    LineZ(b, x - (xs - 1)*dir, y - 1, y - ys + 1, dir, fOff);
  }
  if (fX || fY)
    LineY(b, x, y, y + zs, fOff);
  if (fY || fZ)
    LineX(b, x - (xs - 1)*dir, x - dir, y, fOff);
  if (fX || fZ)
    LineZ(b, x + dir, y - 1, y - ys + 1, dir, fOff);

  // Fill in the three visible faces of the cube.
  if (fY && xs > 1 && zs > 1)
    Block(b, x - (xs - 1)*dir, y + 1, x - dir, y + zs - 1, !ds.fShading);
  if (fZ || fX)
    for (i = 1; i < ys; i++) {
      if (fZ && xs > 1)
        LineX(b, x - (xs - 1 - i)*dir, x + (i - 1)*dir, y - i, !ds.fSkyShade);
      if (fX && zs > 1)
        LineY(b, x + i*dir, y - i + 1, y + zs - 1 - i, fOn);
    }
}


// Draw some or all of the faces of a solid angled orthographic block, with
// one edge oriented toward you. Used by various FillOverview2 functions.

void FillCube2(bitmap b, int x, int y, int xs, int ys, int zs,
  flag fX, flag fY, flag fZ)
{
  int xs2 = xs*ys, i, j;

  // Draw the outline of the cube.
  if (fX) {
    LineY(b, x + xs2 + 1, y - ys - 1, y - ys - 1 + zs, fOff);
    LineZ(b, x + 1, y + zs - 1, y + zs - ys, xs, fOff);
  }
  if (fY) {
    LineY(b, x - xs2 - 1, y - ys - 1, y - ys - 1 + zs, fOff);
    LineZ(b, x - 1, y + zs - 1, y + zs - ys, -xs, fOff);
  }
  if (fZ) {
    LineZ(b, x + xs2, y - ys - 2, y - ys*2 - 1, -xs, fOff);
    LineZ(b, x - xs2, y - ys - 2, y - ys*2 - 1, xs, fOff);
    M0(b, x, y - ys*2 - 2);
  }
  if (fX || fY)
    LineY(b, x, y, y + zs, fOff);
  if (fY || fZ)
    LineZ(b, x - 1, y - 1, y - ys, -xs, fOff);
  if (fX || fZ)
    LineZ(b, x + 1, y - 1, y - ys, xs, fOff);

  // Fill in the three visible faces of the cube.
  if (fZ) {
    LineX(b, x - xs2, x + xs2, y - ys - 1, !ds.fSkyShade);
    for (i = 1; i <= ys; i++) {
      j = (ys - i)*xs;
      LineX(b, x - j, x + j, y - ys - 1 - i, !ds.fSkyShade);
      LineX(b, x - j, x + j, y - ys - 1 + i, !ds.fSkyShade);
    }
  }
  if ((fX || fY) && zs > 2)
    for (i = 0; i < ys; i++) {
      if (fX)
        Block(b, x + i*xs + 1, y - i, x + i*xs + xs, y - i + zs - 2, fOn);
      if (fY)
        Block(b, x - i*xs - 1, y - i, x - i*xs - xs, y - i + zs - 2,
          !ds.fShading);
    }
}


// Draw an orthographic overview of one bitmap within the bounds of another.

void FillOverview(bitmap b2, constbitmap b1)
{
  int x, y, horloc, verloc, dir;

  if (ds.depsiz < 1)
    return;
  dir = ds.fRight*2 - 1;
  for (y = 0; y < MY(b1); y++) {
    verloc = (y + 1)*ds.versiz;
    for (x = ds.fRight ? 0 : MX(b1) - 1; x >= 0 && x < MX(b1); x += dir) {
      if (ds.fRight)
        horloc = (x + 1)*ds.horsiz + (MY(b1) - y - 1)*ds.versiz;
      else
        horloc = x*ds.horsiz + (y + 1)*ds.versiz;
      if (MG(b1, x, y)) {

        // Draw a block for each set pixel.
        FillCube(b2, horloc, verloc, ds.horsiz, ds.versiz, ds.depsiz - 1,
          dir, !Mg(b1, x + dir, y), !Mg(b1, x, y + 1), fTrue);

        // Draw lines removing seams between this and earlier drawn blocks.
        if (ds.fMerge) {
          if (Mg(b1, x - dir, y)) {
            if (ds.versiz > 1)
              LineZ(b2, horloc - (ds.horsiz - 1)*dir, verloc - 1,
                verloc - ds.versiz + 1, dir, !ds.fSkyShade);
            if (ds.depsiz > 2)
              LineY(b2, horloc - ds.horsiz*dir, verloc + 1,
                verloc + ds.depsiz - 2, !ds.fShading);
            if (Mg(b1, x, y - 1) && Mg(b1, x - dir, y - 1))
              MP(b2, horloc - (ds.horsiz - ds.versiz)*dir, verloc - ds.versiz,
                !ds.fSkyShade);
          }
          if (Mg(b1, x, y - 1)) {
            if (ds.horsiz > 1)
              LineX(b2, horloc - (ds.horsiz - ds.versiz - 1)*dir, horloc +
                (ds.versiz - 1)*dir, verloc - ds.versiz, !ds.fSkyShade);
            if (!Mg(b1, x + dir, y - 1) && ds.depsiz > 2)
              LineY(b2, horloc + ds.versiz*dir, verloc - ds.versiz + 1,
                verloc - ds.versiz + ds.depsiz - 2, fOn);
          }
        }
      }
    }
    UpdateDisplay();
  }
}


// Replace a bitmap with an orthographic overview of itself. Implements the
// Render Bitmap Overview command for monochrome 2D bitmaps.

flag DrawOverview(bitmap *b)
{
  bitmap b1 = NULL;

  if (!FBitmapCopy(*b, &b1))
    return fFalse;
  if (!FBitmapSizeSet(b, MX(b1)*ds.horsiz + MY(b1)*ds.versiz + 1,
    MY(b1)*ds.versiz + ds.depsiz))
    {
    DeallocateBitmap(b1);
    return fFalse;
    }
  BitmapSet(*b, !ds.fGroundShade);
  UpdateDisplay();
  FillOverview(*b, b1);
  DeallocateBitmap(b1);
  return fTrue;
}


// Draw some or all of the faces of a solid orthographic block on a color
// bitmap. Used by various color FillOverview functions.

void FillCubeC(colmap c, int x, int y, int xs, int ys, int zs, int dir,
  KV kv0, KV kv1, real rx, real ry, real rz, flag fX, flag fY, flag fZ)
{
  int i;
  KV kvX, kvY, kvZ;

  // Draw the outline of the cube with one color.
  if (fX) {
    LineYC(c, x + ys*dir, y - ys, y - ys + zs, kv0);
    LineZC(c, x + dir, y + zs - 1, y - ys + zs + 1, dir, kv0);
  }
  if (fY) {
    LineXC(c, x - (xs - 1)*dir, x - dir, y + zs, kv0);
    LineYC(c, x - xs*dir, y, y + zs, kv0);
  }
  if (fZ) {
    LineXC(c, x - (xs - ys)*dir, x + (ys - 1)*dir, y - ys, kv0);
    LineZC(c, x - (xs - 1)*dir, y - 1, y - ys + 1, dir, kv0);
  }
  if (fX || fY)
    LineYC(c, x, y, y + zs, kv0);
  if (fY || fZ)
    LineXC(c, x - (xs - 1)*dir, x - dir, y, kv0);
  if (fX || fZ)
    LineZC(c, x + dir, y - 1, y - ys + 1, dir, kv0);

  // Fill in the three visible faces of the cube with another color. Each of
  // the three faces gets shaded with a passed in value.
  kvX = KvShade(kv1, rx); kvY = KvShade(kv1, ry); kvZ = KvShade(kv1, rz);
  if (fY && xs > 1 && zs > 1)
    BlockC(c, x - (xs - 1)*dir, y + 1, x - dir, y + zs - 1, kvY);
  for (i = 1; i < ys; i++) {
    if (fZ && xs > 1)
      LineXC(c, x - (xs - 1 - i)*dir, x + (i - 1)*dir, y - i, kvZ);
    if (fX && zs > 1)
      LineYC(c, x + i*dir, y - i + 1, y + zs - 1 - i, kvX);
  }
}


// Draw an orthographic overview of a monochrome bitmap within the bounds of a
// color bitmap. The color of each block comes from a second color bitmap.

void FillOverviewC(colmap c, constbitmap b, constcolmap c2)
{
  int x, y, horloc, verloc, dir;
  real rx = 0.0, ry = 0.0, rz = 0.0;
  KV kv, kvT;

  if (ds.depsiz < 1)
    return;
  if (ds.fShading) {
    rx = ds.vLight.x / 100.0;
    ry = ds.vLight.y / 100.0;
    rz = ds.vLight.z / 100.0;
  }
  dir = ds.fRight*2 - 1;
  for (y = 0; y < MY(b); y++) {
    verloc = (y + 1)*ds.versiz;
    for (x = ds.fRight ? 0 : MX(b) - 1; x >= 0 && x < MX(b); x += dir) {
      if (ds.fRight)
        horloc = (x + 1)*ds.horsiz + (MY(b) - y - 1)*ds.versiz;
      else
        horloc = x*ds.horsiz + (y + 1)*ds.versiz;
      if (MG(b, x, y)) {

        // Draw a colored block for each set pixel.
        kv = Cg(c2, x, y);
        FillCubeC(c, horloc, verloc, ds.horsiz, ds.versiz, ds.depsiz - 1, dir,
          ds.kvTrim, kv, rx, ry, rz,
          !Mg(b, x + dir, y), !Mg(b, x, y + 1), fTrue);

        // Draw lines removing seams between this and earlier drawn blocks, in
        // a color half way between the color used for the two faces.
        if (ds.fMerge) {
          if (Mg(b, x - dir, y)) {
            kvT = KvBlend(kv, Cg(c2, x - dir, y));
            if (ds.versiz > 1)
              LineZC(c, horloc - (ds.horsiz - 1)*dir, verloc - 1,
                verloc - ds.versiz + 1, dir, KvShade(kvT, rz));
            if (ds.depsiz > 2)
              LineYC(c, horloc - ds.horsiz*dir, verloc + 1,
                verloc + ds.depsiz - 2, KvShade(kvT, ry));
            if (Mg(b, x, y - 1) && Mg(b, x - dir, y - 1))
              CP(c, horloc - (ds.horsiz - ds.versiz)*dir, verloc - ds.versiz,
                KvShade(KvBlend(kvT, KvBlend(Cg(c2, x, y - 1),
                Cg(c2, x - dir, y - 1))), rz));
          }
          if (Mg(b, x, y - 1)) {
            kvT = KvBlend(kv, Cg(c2, x, y - 1));
            if (ds.horsiz > 1)
              LineXC(c, horloc - (ds.horsiz - ds.versiz - 1)*dir, horloc +
                (ds.versiz - 1)*dir, verloc - ds.versiz, KvShade(kvT, rz));
            if (!Mg(b, x + dir, y - 1) && ds.depsiz > 2)
              LineYC(c, horloc + ds.versiz*dir, verloc - ds.versiz + 1,
                verloc - ds.versiz + ds.depsiz - 2, KvShade(kvT, rx));
          }
        }
      }
    }
    UpdateDisplay();
  }
}


// Replace a color bitmap with an orthographic overview of a monochrome
// bitmap, where the color of each block comes from the old contents of that
// bitmap. Implements the Render Bitmap Overview command for color 2D bitmaps.

flag DrawOverviewC(colmap *c, constbitmap b)
{
  colmap c1 = NULL;

  if (!FColmapCopy(*c, &c1))
    return fFalse;
  if (!FColmapSizeSet(c, MX(b)*ds.horsiz + MY(b)*ds.versiz + 1,
    MY(b)*ds.versiz + ds.depsiz))
    {
    DeallocateColmap(c1);
    return fFalse;
    }
  ColmapSet(*c, ds.kvGroundLo);
  UpdateDisplay();
  FillOverviewC(*c, b, c1);
  DeallocateColmap(c1);
  return fTrue;
}


// Draw an orthographic overview of a 3D bitmap within the bounds of another
// bitmap.

void FillOverviewCube(bitmap b, constcubemap t, int lensiz, int widsiz,
  int depsiz, flag fView, flag fMerge)
{
  int x, y, z, horloc, verloc, dir;
  flag f1, f2;

  dir = fView*2 - 1;
  for (y = 0; y < TY(t); y++) {
    for (z = TZ(t)-1; z >= 0; z--) {
      verloc = z*depsiz + (y + 1)*widsiz;
      for (x = fView ? 0 : TX(t) - 1; x >= 0 && x < TX(t); x += dir) {
        if (fView)
          horloc = (x + 1)*lensiz + (TY(t) - 1 - y)*widsiz;
        else
          horloc = x*lensiz + (y + 1)*widsiz;
        if (TG(t, x, y, z)) {

          // Draw a block for each set pixel.
          FillCube(b, horloc, verloc, lensiz, widsiz, depsiz, dir,
            !Tg(t, x + dir, y, z), !Tg(t, x, y + 1, z), !Tg(t, x, y, z - 1));

          // Draw lines removing seams between this and earlier drawn blocks.
          if (fMerge) {
            if (Tg(t, x - dir, y, z)) {
              if (depsiz > 1)
                LineY(b, horloc - lensiz*dir, verloc + 1, verloc + depsiz - 1,
                  !ds.fShading);
              if (widsiz > 1)
                LineZ(b, horloc - (lensiz - 1)*dir, verloc - 1,
                  verloc - widsiz + 1, dir, !ds.fSkyShade);
              f1 = Tg(t, x, y, z + 1) && Tg(t, x - dir, y, z + 1);
              if (f1)
                MP(b, horloc - lensiz*dir, verloc + depsiz, !ds.fShading);
              f1 = Tg(t, x, y - 1, z) && Tg(t, x - dir, y - 1, z);
              f2 = !(Tg(t, x, y - 1, z - 1) || Tg(t, x - dir, y - 1, z - 1));
              if (f1 && f2)
                MP(b, horloc - (lensiz - widsiz)*dir, verloc - widsiz,
                  !ds.fSkyShade);
            }
            if (Tg(t, x, y, z + 1)) {
              if (lensiz > 1)
                LineX(b, horloc - (lensiz - 1)*dir, horloc - dir,
                  verloc + depsiz, !ds.fShading);
              if (!Tg(t, x + dir, y, z + 1) && widsiz > 1)
                LineZ(b, horloc + dir, verloc + depsiz - 1,
                  verloc - widsiz + depsiz + 1, dir, fOn);
              f1 = Tg(t, x, y - 1, z) && Tg(t, x, y - 1, z + 1);
              f2 = !(Tg(t, x + dir, y - 1, z) || Tg(t, x + dir, y, z + 1));
              if (f1 && f2 && !Tg(t, x + dir, y - 1, z + 1))
                M1(b, horloc + widsiz*dir, verloc - widsiz + depsiz);
            }
            if (Tg(t, x, y - 1, z)) {
              if (!Tg(t, x + dir, y - 1, z) && depsiz > 1)
                LineY(b, horloc + widsiz*dir, verloc - widsiz + 1,
                  verloc - widsiz + depsiz - 1, fOn);
              if (!Tg(t, x, y - 1, z - 1) && lensiz > 1)
                LineX(b, horloc - (lensiz - widsiz - 1)*dir,
                  horloc + (widsiz - 1)*dir, verloc - widsiz, !ds.fSkyShade);
            }
          }
        }
      }
    }
    UpdateDisplay();
  }
}


// Replace a 3D bitmap with an orthographic overview of itself. Implements the
// Render Bitmap Overview command for monochrome 3D bitmaps.

flag DrawOverviewCube(bitmap *b)
{
  cubemap t = NULL;

  if (!FBitmapCopy(*b, &t))
    return fFalse;
  if (!FBitmapSizeSet(b, TX(t)*ds.horsiz + TY(t)*ds.versiz + 1,
    TZ(t)*ds.depsiz + TY(t)*ds.versiz + 1))
    {
    DeallocateCube(t);
    return fFalse;
    }
  BitmapSet(*b, !ds.fGroundShade);
  UpdateDisplay();
  FillOverviewCube(*b, t, ds.horsiz, ds.versiz, ds.depsiz,
    ds.fRight, ds.fMerge);
  DeallocateCube(t);
  return fTrue;
}


// Draw an orthographic overview of a 3D bitmap within the bounds of a color
// bitmap. The color of each block comes from a second color bitmap.

void FillOverviewCubeC(colmap c, constcubemap t, constcolmap c2,
  int lensiz, int widsiz, int depsiz, flag fView, flag fMerge)
{
  int x, y, z, horloc, verloc, dir;
  real rx = 0.0, ry = 0.0, rz = 0.0;
  KV kv, kvT;
  flag f1, f2;

  if (ds.fShading) {
    rx = ds.vLight.x / 100.0;
    ry = ds.vLight.y / 100.0;
    rz = ds.vLight.z / 100.0;
  }
  dir = fView*2 - 1;
  for (y = 0; y < TY(t); y++) {
    for (z = TZ(t)-1; z >= 0; z--) {
      verloc = z*depsiz + (y + 1)*widsiz;
      for (x = fView ? 0 : TX(t) - 1; x >= 0 && x < TX(t); x += dir) {
        if (fView)
          horloc = (x + 1)*lensiz + (TY(t) - 1 - y)*widsiz;
        else
          horloc = x*lensiz + (y + 1)*widsiz;
        if (TG(t, x, y, z)) {

          // Draw a colored block for each set pixel.
          kv = Cg3(c2, x, y, z);
          FillCubeC(c, horloc, verloc, lensiz, widsiz, depsiz, dir,
            ds.kvTrim, kv, rx, ry, rz,
            !Tg(t, x + dir, y, z), !Tg(t, x, y + 1, z), !Tg(t, x, y, z - 1));

          // Draw lines removing seams between this and earlier drawn blocks,
          // in a color half way between the color used for the two faces.
          if (fMerge) {
            if (Tg(t, x - dir, y, z)) {
              kvT = KvBlend(kv, Cg3(c2, x - dir, y, z));
              if (depsiz > 1)
                LineYC(c, horloc - lensiz*dir, verloc + 1,
                  verloc + depsiz - 1, KvShade(kvT, ry));
              if (widsiz > 1)
                LineZC(c, horloc - (lensiz - 1)*dir, verloc - 1,
                  verloc - widsiz + 1, dir, KvShade(kvT, rz));
              f1 = Tg(t, x, y, z + 1) && Tg(t, x - dir, y, z + 1);
              if (f1)
                CP(c, horloc - lensiz*dir, verloc + depsiz,
                  KvShade(KvBlend(kvT, KvBlend(Cg3(c2, x, y, z + 1),
                  Cg3(c2, x - dir, y, z + 1))), ry));
              f1 = Tg(t, x, y - 1, z) && Tg(t, x - dir, y - 1, z);
              f2 = !(Tg(t, x, y - 1, z - 1) || Tg(t, x - dir, y - 1, z - 1));
              if (f1 && f2)
                CP(c, horloc - (lensiz - widsiz)*dir, verloc - widsiz,
                  KvShade(KvBlend(kvT, KvBlend(Cg3(c2, x, y - 1, z),
                  Cg3(c2, x - dir, y - 1, z))), rz));
            }
            if (Tg(t, x, y, z + 1)) {
              kvT = KvBlend(kv, Cg3(c2, x, y, z + 1));
              if (lensiz > 1)
                LineXC(c, horloc - (lensiz - 1)*dir, horloc - dir,
                  verloc + depsiz, KvShade(kvT, ry));
              if (!Tg(t, x + dir, y, z + 1) && widsiz > 1)
                LineZC(c, horloc + dir, verloc + depsiz - 1,
                  verloc - widsiz + depsiz + 1, dir, KvShade(kvT, rx));
              f1 = Tg(t, x, y - 1, z) && Tg(t, x, y - 1, z + 1);
              f2 = !(Tg(t, x + dir, y - 1, z) || Tg(t, x + dir, y, z + 1));
              if (f1 && f2 && !Tg(t, x + dir, y - 1, z + 1))
                CP(c, horloc + widsiz*dir, verloc - widsiz + depsiz,
                  KvShade(KvBlend(kvT, KvBlend(Cg3(c2, x, y - 1, z),
                  Cg3(c2, x, y - 1, z + 1))), rx));
            }
            if (Tg(t, x, y - 1, z)) {
              kvT = KvBlend(kv, Cg3(c2, x, y - 1, z));
              if (!Tg(t, x + dir, y - 1, z) && depsiz > 1)
                LineYC(c, horloc + widsiz*dir, verloc - widsiz + 1,
                  verloc - widsiz + depsiz - 1, KvShade(kvT, rx));
              if (!Tg(t, x, y - 1, z - 1) && lensiz > 1)
                LineXC(c, horloc - (lensiz - widsiz - 1)*dir, horloc +
                  (widsiz - 1)*dir, verloc - widsiz, KvShade(kvT, rz));
            }
          }
        }
      }
    }
    UpdateDisplay();
  }
}


// Replace a color bitmap with an orthographic overview of a 3D monochrome
// bitmap, where the color of each block comes from the old contents of that
// bitmap. Implements the Render Bitmap Overview command for color 3D bitmaps.

flag DrawOverviewCubeC(colmap *c, constcubemap t)
{
  colmap c1 = NULL;

  if (!FColmapCopy(*c, &c1))
    return fFalse;
  if (!FColmapSizeSet(c, TX(t)*ds.horsiz + TY(t)*ds.versiz + 1,
    TZ(t)*ds.depsiz + TY(t)*ds.versiz + 1))
    {
    DeallocateColmap(c1);
    return fFalse;
    }
  ColmapSet(*c, ds.kvGroundLo);
  UpdateDisplay();
  FillOverviewCubeC(*c, t, c1, ds.horsiz, ds.versiz, ds.depsiz,
    ds.fRight, ds.fMerge);
  DeallocateColmap(c1);
  return fTrue;
}


// Draw an angled orthographic overview of one bitmap within the bounds of
// another.

void FillOverview2(bitmap b2, constbitmap b1)
{
  int xs = ds.horsiz*ds.versiz, x, y, horloc, verloc;

  if (ds.depsiz < 1)
    return;
  for (y = 0; y < MY(b1); y++) {
    for (x = 0; x < MX(b1); x++) if (MG(b1, x, y)) {
      horloc = (x + (MY(b1)-y)) * (xs + 1);
      verloc = (x + y + 2)*(ds.versiz + 1);

      // Draw a block for each set pixel.
      FillCube2(b2, horloc, verloc, ds.horsiz, ds.versiz, ds.depsiz - 1,
        !Mg(b1, x + 1, y), !Mg(b1, x, y + 1), fTrue);

      // Draw lines removing seams between this and earlier drawn blocks.
      if (ds.fMerge) {
        if (Mg(b1, x - 1, y)) {
          LineZ(b2, horloc - xs, verloc - ds.versiz - 2,
            verloc - ds.versiz*2 - 1, ds.horsiz, !ds.fSkyShade);
          if (ds.depsiz > 2)
            LineY(b2, horloc - xs - 1, verloc - ds.versiz,
              verloc - ds.versiz + ds.depsiz - 3, !ds.fShading);
          if (Mg(b1, x, y - 1) && Mg(b1, x - 1, y - 1))
            MP(b2, horloc, verloc - ds.versiz*2 - 2, !ds.fSkyShade);
        }
        if (Mg(b1, x, y - 1)) {
          LineZ(b2, horloc + xs, verloc - ds.versiz - 2,
            verloc - ds.versiz*2 - 1, -ds.horsiz, !ds.fSkyShade);
          if (!Mg(b1, x + 1, y - 1) && ds.depsiz > 2)
            LineY(b2, horloc + xs + 1, verloc - ds.versiz,
              verloc - ds.versiz + ds.depsiz - 3, fOn);
        }
      }
    }
    UpdateDisplay();
  }
}


// Replace a bitmap with an angled orthographic overview of itself. Implements
// the Overview2 operation for monochrome 2D bitmaps.

flag DrawOverview2(bitmap *b)
{
  bitmap b1 = NULL;

  if (!FBitmapCopy(*b, &b1))
    return fFalse;
  if (!FBitmapSizeSet(b, (MX(b1) + MY(b1))*(ds.horsiz*ds.versiz + 1) + 1,
    (MX(b1) + MY(b1))*(ds.versiz + 1) + ds.depsiz))
    {
    DeallocateBitmap(b1);
    return fFalse;
    }
  BitmapSet(*b, !ds.fGroundShade);
  UpdateDisplay();
  FillOverview2(*b, b1);
  DeallocateBitmap(b1);
  return fTrue;
}


// Draw an angled orthographic overview of a 3D bitmap within the bounds of
// another bitmap.

void FillOverviewCube2(bitmap b, constcubemap t,
  int lensiz, int widsiz, int depsiz, flag fMerge)
{
  int xs = ds.horsiz*ds.versiz, x, y, z, horloc, verloc;
  flag f1, f2;

  for (y = 0; y < TY(t); y++) {
    for (z = TZ(t)-1; z >= 0; z--)
      for (x = 0; x < TX(t); x++)
        if (TG(t, x, y, z)) {
          horloc = (x + (TY(t)-y)) * (xs + 1);
          verloc = (x + y + 2)*(ds.versiz + 1) + z*ds.depsiz;

          // Draw a block for each set pixel.
          FillCube2(b, horloc, verloc, lensiz, widsiz, depsiz,
            !Tg(t, x + 1, y, z), !Tg(t, x, y + 1, z), !Tg(t, x, y, z - 1));

          // Draw lines removing seams between this and earlier drawn blocks.
          if (fMerge) {
            if (Tg(t, x - 1, y, z)) {
              if (depsiz > 1)
                LineY(b, horloc - xs - 1, verloc - widsiz,
                  verloc - widsiz + depsiz - 2, !ds.fShading);
              if (widsiz > 0)
                LineZ(b, horloc - xs, verloc - widsiz - 2,
                  verloc - widsiz*2 - 1, lensiz, !ds.fSkyShade);
              f1 = Tg(t, x, y, z + 1) && Tg(t, x - 1, y, z + 1);
              if (f1)
                MP(b, horloc - xs - 1, verloc - widsiz + depsiz - 1,
                  !ds.fShading);
              f1 = Tg(t, x, y - 1, z) && Tg(t, x - 1, y - 1, z);
              f2 = !(Tg(t, x, y - 1, z - 1) || Tg(t, x - 1, y - 1, z - 1));
              if (f1 && f2)
                MP(b, horloc, verloc - widsiz*2 - 2, !ds.fSkyShade);
            }
            if (Tg(t, x, y, z + 1)) {
              if (lensiz > 0)
                LineZ(b, horloc - 1, verloc + depsiz - 1,
                  verloc + depsiz - widsiz, -lensiz, !ds.fShading);
              if (!Tg(t, x + 1, y, z + 1) && widsiz > 0)
                LineZ(b, horloc + 1, verloc + depsiz - 1,
                  verloc + depsiz - widsiz, lensiz, fOn);
              f1 = Tg(t, x, y - 1, z) && Tg(t, x, y - 1, z + 1);
              f2 = !(Tg(t, x + 1, y - 1, z) || Tg(t, x + 1, y, z + 1));
              if (f1 && f2 && !Tg(t, x + 1, y - 1, z + 1))
                M1(b, horloc + xs + 1, verloc - widsiz + depsiz - 1);
            }
            if (Tg(t, x, y - 1, z)) {
              if (!Tg(t, x + 1, y - 1, z) && depsiz > 1)
                LineY(b, horloc + xs + 1, verloc - widsiz,
                  verloc - widsiz + depsiz - 2, fOn);
              if (!Tg(t, x, y - 1, z - 1) && lensiz > 0)
                LineZ(b, horloc + xs, verloc - widsiz - 2,
                  verloc - widsiz*2 - 1, -lensiz, !ds.fSkyShade);
            }
          }
        }
    UpdateDisplay();
  }
}


// Replace a 3D bitmap with an angled orthographic overview of itself.
// Implements the Overview2 operation for monochrome 3D bitmaps.

flag DrawOverviewCube2(bitmap *b)
{
  cubemap t = NULL;

  if (!FBitmapCopy(*b, &t))
    return fFalse;
  if (!FBitmapSizeSet(b, (TX(t) + TY(t))*(ds.horsiz*ds.versiz + 1) + 1,
    (TX(t) + TY(t))*(ds.versiz + 1) + TZ(t)*ds.depsiz + 1))
    {
    DeallocateCube(t);
    return fFalse;
    }
  BitmapSet(*b, !ds.fGroundShade);
  UpdateDisplay();
  FillOverviewCube2(*b, t, ds.horsiz, ds.versiz, ds.depsiz, ds.fMerge);
  DeallocateCube(t);
  return fTrue;
}


#define ZAltitude(kv) (NMin(NMultDiv(kv, zscale, 1000), zmax) - 1)

// Draw an orthographic overview of a color bitmap, where the color value of
// each pixel indicates the height of that block, within the bounds of a
// monochrome bitmap.

void FillOverviewAltitude(bitmap b, constcolmap c, int zscale, int zmax)
{
  int x, y, z, z1, z2, horloc, verloc, dir;
  KV kvT;

  dir = ds.fRight*2 - 1;
  for (y = 0; y < CY(c); y++) {
    verloc = (y + 1)*ds.versiz + zmax - 1;
    for (x = ds.fRight ? 0 : CX(c) - 1; x >= 0 && x < CX(c); x += dir) {
      if (ds.fRight)
        horloc = (x + 1)*ds.horsiz + (CY(c) - y - 1)*ds.versiz;
      else
        horloc = x*ds.horsiz + (y + 1)*ds.versiz;
      kvT = CG(c, x, y);
      z = ZAltitude(kvT);
      if (z >= 0) {

        // Draw a block of appropriate height for each non-zero pixel.
        FillCube(b, horloc, verloc - z, ds.horsiz, ds.versiz, z, dir,
          Cg(c, x + dir, y) < kvT, Cg(c, x, y + 1) < kvT, fTrue);

        // Draw lines removing seams between this and earlier drawn blocks.
        if (ds.fMerge) {
          z1 = ZAltitude(Cg(c, x - dir, y));
          if (z1 >= 0) {
            if (z == z1 && ds.versiz > 1)
              LineZ(b, horloc - (ds.horsiz - 1)*dir, verloc - z - 1,
                verloc - z - ds.versiz + 1, dir, !ds.fSkyShade);
            z2 = Min(z1, z);
            if (z2 > 1)
              LineY(b, horloc - ds.horsiz*dir, verloc - z2 + 1, verloc - 1,
                !ds.fShading);
            if (z == z1 && z == ZAltitude(Cg(c, x, y - 1)) &&
              z == ZAltitude(Cg(c, x - dir, y - 1)))
              MP(b, horloc - (ds.horsiz - ds.versiz)*dir,
                verloc - z - ds.versiz, !ds.fSkyShade);
          }
          z1 = ZAltitude(Cg(c, x, y - 1));
          if (z1 >= 0) {
            if (z == z1 && ds.horsiz > 1)
              LineX(b, horloc - (ds.horsiz - ds.versiz - 1)*dir, horloc +
                (ds.versiz - 1)*dir, verloc - z - ds.versiz, !ds.fSkyShade);
            z1 = Min(z1, z);
            if (z1 > 1) {
              z2 = ZAltitude(Cg(c, x + dir, y - 1)); z2 = Max(z2, 0);
              if (z1 > z2 + 1)
                LineY(b, horloc + ds.versiz*dir, verloc - ds.versiz - z1 + 1,
                  verloc - ds.versiz - z2 - 1, fOn);
            }
          }
        }
      }
    }
    UpdateDisplay();
  }
}


// Draw an orthographic overview of a color bitmap, where the color value of
// each pixel indicates the height of that block, in a new monochrome bitmap.
// Implements the Altitude operation for monochrome bitmaps.

flag DrawOverviewAltitude(bitmap *b, constcolmap c, int zscale, int zmax)
{
  long l;

  if (zscale >= 0)
    zmax = NMultDiv(zmax, zscale, 1000);
  else {
    l = ColmapFind(c, 0);
    zscale = (zmax * 1000 + 499) / (l == 0 ? 1 : l);
  }
  if (!FBitmapSizeSet(b, CX(c)*ds.horsiz + CY(c)*ds.versiz + 1,
    CY(c)*ds.versiz + zmax))
    return fFalse;
  BitmapSet(*b, !ds.fGroundShade);
  UpdateDisplay();
  FillOverviewAltitude(*b, c, zscale, zmax);
  return fTrue;
}


// Draw an orthographic overview of a color bitmap, where the color value of
// each pixel indicates the height of that block, within the bounds of another
// color bitmap. The color to use for each block comes from a third bitmap.

void FillOverviewAltitudeC(colmap c, constcolmap c1, constcolmap c2,
  int zscale, int zmax)
{
  int x, y, z, z1, z2, horloc, verloc, dir;
  real rx = 0.0, ry = 0.0, rz = 0.0;
  KV kv, kvT;

  if (ds.fShading) {
    rx = ds.vLight.x / 100.0;
    ry = ds.vLight.y / 100.0;
    rz = ds.vLight.z / 100.0;
  }
  dir = ds.fRight*2 - 1;
  for (y = 0; y < CY(c1); y++) {
    verloc = (y + 1)*ds.versiz + zmax - 1;
    for (x = ds.fRight ? 0 : CX(c1) - 1; x >= 0 && x < CX(c1); x += dir) {
      if (ds.fRight)
        horloc = (x + 1)*ds.horsiz + (CY(c1) - y - 1)*ds.versiz;
      else
        horloc = x*ds.horsiz + (y + 1)*ds.versiz;
      kvT = CG(c1, x, y);
      z = ZAltitude(kvT);
      if (z >= 0) {

        // Draw a colored block of appropriate height for each non-zero pixel.
        kv = Cg(c2, x, y);
        FillCubeC(c, horloc, verloc - z, ds.horsiz, ds.versiz, z, dir,
          ds.kvTrim, kv, rx, ry, rz,
          Cg(c1, x + dir, y) < kvT, Cg(c1, x, y + 1) < kvT, fTrue);

        // Draw lines removing seams between this and earlier drawn blocks, in
        // a color half way between the color used for the two faces.
        if (ds.fMerge) {
          z1 = ZAltitude(Cg(c1, x - dir, y));
          if (z1 >= 0) {
            kvT = KvBlend(kv, Cg(c2, x - dir, y));
            if (z == z1 && ds.versiz > 1)
              LineZC(c, horloc - (ds.horsiz - 1)*dir, verloc - z - 1,
                verloc - z - ds.versiz + 1, dir, KvShade(kvT, rz));
            z2 = Min(z1, z);
            if (z2 > 1)
              LineYC(c, horloc - ds.horsiz*dir, verloc - z2 + 1, verloc - 1,
                KvShade(kvT, ry));
            if (z == z1 && z == ZAltitude(Cg(c1, x, y - 1)) &&
              z == ZAltitude(Cg(c1, x - dir, y - 1)))
              CP(c, horloc - (ds.horsiz - ds.versiz)*dir,
                verloc - z - ds.versiz, KvShade(KvBlend(kvT,
                KvBlend(Cg(c2, x, y - 1), Cg(c2, x - dir, y - 1))), rz));
          }
          z1 = ZAltitude(Cg(c1, x, y - 1));
          if (z1 >= 0) {
            kvT = KvBlend(kv, Cg(c2, x, y - 1));
            if (z == z1 && ds.horsiz > 1)
              LineXC(c, horloc - (ds.horsiz - ds.versiz - 1)*dir,
                horloc + (ds.versiz - 1)*dir, verloc - z - ds.versiz,
                KvShade(kvT, rz));
            z1 = Min(z1, z);
            if (z1 > 1) {
              z2 = ZAltitude(Cg(c1, x + dir, y - 1)); z2 = Max(z2, 0);
              if (z1 > z2 + 1)
                LineYC(c, horloc + ds.versiz*dir, verloc - ds.versiz - z1 + 1,
                  verloc - ds.versiz - z2 - 1, KvShade(kvT, rx));
            }
          }
        }
      }
    }
    UpdateDisplay();
  }
}


// Replace a color bitmap with an orthographic overview of itself, where the
// color value of each pixel indicates the height of that block. The color to
// use for each block comes from a second bitmap. Implements the Altitude
// operation for color bitmaps.

flag DrawOverviewAltitudeC(colmap *c, constcolmap c2, int zscale, int zmax)
{
  colmap c1 = NULL;
  long l;

  if (!FColmapCopy(*c, &c1))
    return fFalse;
  if (zscale >= 0)
    zmax = NMultDiv(zmax, zscale, 1000);
  else {
    l = ColmapFind(c1, 0);
    zscale = (zmax * 1000 + 499) / (l == 0 ? 1 : l);
  }
  if (!FColmapSizeSet(c, CX(c1)*ds.horsiz + CY(c1)*ds.versiz + 1,
    CY(c1)*ds.versiz + zmax))
    {
    DeallocateColmap(c1);
    return fFalse;
    }
  ColmapSet(*c, ds.kvGroundLo);
  UpdateDisplay();
  FillOverviewAltitudeC(*c, c1, c2 != NULL ? c2 : c1, zscale, zmax);
  DeallocateColmap(c1);
  return fTrue;
}


/*
******************************************************************************
** Render Pyramid
******************************************************************************
*/

// Return whether a point between four blocks forms a visible corner that
// should have an edge line drawn at that point.

flag FCorner(constbitmap b, int x, int y)
{
  flag f0, f1, f2, f3;
  int count;

  // A single block forms a convex corner, three blocks form a concave corner,
  // and two opposite blocks form a checkerboard corner.

  f0 = Mg(b, x, y);
  f1 = Mg(b, x - 1, y);
  f2 = Mg(b, x, y - 1);
  f3 = Mg(b, x - 1, y - 1);
  count = f0 + f1 + f2 + f3;
  return FOdd(count) || (count == 2 && f0 == f3);
}


// Draw an edge line from an exposed block corner to the vanishing point. Stop
// drawing if the line becomes no longer visible behind other blocks.

void PLine(bitmap b2, colmap c, constbitmap b1, int x1, int y1,
  int x2, int y2, int xv, int yv)
{
  int x = x1, y = y1, dx = x2 - x1, dy = y2 - y1, xInc, yInc, xInc2, yInc2,
    d, dInc, z, zMax, xT, yT;
  flag fColor = (c != NULL), fStarted = fFalse, fProbation = fFalse,
    f0, f1, f2, f3;

  // If blocks are transparent, just draw the whole line.
  if (!ds.fMerge) {
    if (!fColor)
      Line(b2, x1, y1, x2, y2, fOff);
    else
      LineC(c, x1, y1, x2, y2, ds.kvTrim);
    return;
  }

  // If vanishing point is covered by a block, don't draw edges of that block.
  xT = x1 / ds.horsiz; yT = y1 / ds.versiz;
  if (Mg(b1, xT - (xT > xv), yT - (yT > yv)))
    return;

  // Determine slope.
  if (NAbs(dx) >= NAbs(dy)) {
    xInc = NSgn(dx); yInc = 0;
    xInc2 = 0; yInc2 = NSgn(dy);
    zMax = NAbs(dx); dInc = NAbs(dy);
    d = zMax - (!FOdd(dx) && x1 > x2);
  } else {
    xInc = 0; yInc = NSgn(dy);
    xInc2 = NSgn(dx); yInc2 = 0;
    zMax = NAbs(dy); dInc = NAbs(dx);
    d = zMax - (!FOdd(dy) && y1 > y2);
  }
  d >>= 1;

  // Loop over long axis, adjusting short axis for slope as needed.
  for (z = 0; z <= zMax; z++) {
    f0 = Mg(b1, x / ds.horsiz, y / ds.versiz);
    f1 = Mg(b1, (x-1) / ds.horsiz, y / ds.versiz);
    f2 = Mg(b1, x / ds.horsiz, (y-1) / ds.versiz);
    f3 = Mg(b1, (x-1) / ds.horsiz, (y-1) / ds.versiz);

    // If entirely within a block, definitely stop drawing.
    if (f0 && f1 && f2 && f3)
      break;

    // If adjacent to a block, that's ok unless it happens twice in a row.
    if (f0 || f1 || f2 || f3) {
      if (fStarted) {
        if (fProbation)
          break;
        fProbation = fTrue;
      }
    } else {
      fStarted = fTrue;
      fProbation = fFalse;
    }

    // Not within a block, so draw next pixel of line.
    if (!fColor)
      M0(b2, x, y);
    else
      CP(c, x, y, ds.kvTrim);
    x += xInc; y += yInc; d += dInc;
    if (d >= zMax) {
      x += xInc2; y += yInc2; d -= zMax;
    }
  }
}


// Draw a single block in a single point perspective graphic.

void FillBlock(bitmap b2, colmap c, constbitmap b1, int x, int y,
  int xv, int yv)
{
  int x1, y1, x2, y2;
  flag fColor = (c != NULL);

  if (!Mg(b1, x, y) || !ds.fMerge)
    return;

  // Draw the front face, and the faces that go to the vanishing point.

  if (!fColor) {
    if ((x < xv && !Mg(b1, x+1, y)) || (x > xv && !Mg(b1, x-1, y)) ||
      (y < yv && !Mg(b1, x, y+1)) || (y > yv && !Mg(b1, x, y-1))) {
      x1 = (x + (y == yv && x < xv))*ds.horsiz;
      y1 = (y + ((y < yv && x <= xv) || (y > yv && x > xv)))*ds.versiz;
      x2 = (x + (y != yv || x < xv))*ds.horsiz;
      y2 = (y + ((y >= yv || x >= xv) && (y <= yv || x < xv)))*ds.versiz;
      FTriangle(b2, ds.horv, ds.verv, x1, y1, x2, y2, fOn);
    }
    Block(b2, x*ds.horsiz, y*ds.versiz, (x+1)*ds.horsiz, (y+1)*ds.versiz,
      !ds.fShading);
  } else {
    x1 = x*ds.horsiz; y1  = y*ds.versiz;
    x2 = x1 + ds.horsiz; y2 = y1 + ds.versiz;
    if (x < xv && !Mg(b1, x+1, y))
      FTriangleC(c, ds.horv, ds.verv, x2, y1, x2, y2,
        ds.fShading ? ds.kvGroundLo : ds.kvSkyHi, 0);
    if (x > xv && !Mg(b1, x-1, y))
      FTriangleC(c, ds.horv, ds.verv, x1, y1, x1, y2,
        ds.fShading ? ds.kvGroundHi : ds.kvSkyHi, 0);
    if (y < yv && !Mg(b1, x, y+1))
      FTriangleC(c, ds.horv, ds.verv, x1, y2, x2, y2,
        ds.fShading ? ds.kvSkyLo    : ds.kvSkyHi, 0);
    if (y > yv && !Mg(b1, x, y-1))
      FTriangleC(c, ds.horv, ds.verv, x1, y1, x2, y1,
        ds.kvSkyHi, 0);
    BlockC(c, x*ds.horsiz, y*ds.versiz, (x+1)*ds.horsiz, (y+1)*ds.versiz,
      ds.kvObject);
  }
}


// Draw a line of blocks in a single point perspective graphic, in the given
// order. It's assumed the vanishing point doesn't intersect the line.

void FillLine(bitmap b2, colmap c, constbitmap b1, int x1, int y1,
  int x2, int y2, int xv, int yv)
{
  int xi, yi, x, y;

  // Draw blocks in a line from (x1, y1) to (x2, y2).
  Assert(x1 == x2 || y1 == y2);
  if (x1 == x2) {
    yi = y1 < y2 ? 1 : -1; xi = 0;
  } else {
    xi = x1 < x2 ? 1 : -1; yi = 0;
  }
  x = x1; y = y1;
  loop {
    FillBlock(b2, c, b1, x, y, xv, yv);
    if (x == x2 && y == y2)
      break;
    x += xi; y += yi;
  }
}


// Draw a grid of blocks in a single point perspective graphic, in the given
// order from one corner to the opposite. It's assumed the vanishing point
// doesn't intersect the grid. Lines of blocks are drawn touching the diagonal
// leading up to the end corner. For example, a 5x4 grid from lower left to
// upper right has blocks drawn in the following order:
//
// 16 17 18 19 20
//  9 10 11 12 15
//  4  5  6  8 14
//  1  2  3  7 13

void FillSection(bitmap b2, colmap c, constbitmap b1,
  int x1, int y1, int x2, int y2, int xv, int yv)
{
  int xi, yi, xd, yd, xo, yo, iMax, i;

  // Draw blocks in a grid from (x1, y1) to (x2, y2).
  if (x2 < 0 || y2 < 0)
    return;
  xi = x2 > x1 ? 1 : -1; yi = y2 > y1 ? 1 : -1;
  xd = NAbs(x2 - x1) + 1; yd = NAbs(y2 - y1) + 1;
  if (xd > yd) {
    xo = xd - yd; yo = 0;
  } else {
    yo = yd - xd; xo = 0;
  }

  // Draw the first line touching the start corner.
  FillLine(b2, c, b1, x1, y1, x1 + xo*xi, y1 + yo*yi, xv, yv);

  // Draw pairs of lines growing out from the first line like layers of an
  // onion, until the final layer forming the edges touching the end corner.
  iMax = xd < yd ? xd : yd;
  for (i = 1; i < iMax; i++) {
    FillLine(b2, c, b1,
      x1 + (xo+i)*xi, y1, x1 + (xo+i)*xi, y1 + (yo+i-1)*yi, xv, yv);
    FillLine(b2, c, b1,
      x1, y1 + (yo+i)*yi, x1 + (xo+i)*xi, y1 + (yo+i)*yi, xv, yv);
  }
}


// Draw a single point perspective graphic of one bitmap in another. Each
// pixel becomes a rectangle with lines leading to the vanishing point.

void FillPyramid(bitmap b2, colmap c, constbitmap b1)
{
  int xv, yv, x, y;
  flag fColor = (c != NULL);

  xv = ds.horv / ds.horsiz;
  yv = ds.verv / ds.versiz;
  if (xv > MX(b1))
    xv = MX(b1);
  if (yv > MY(b1))
    yv = MY(b1);

  // Draw the blocks in four quadrants, divided by the location of the
  // vanishing point.

  FillSection(b2, c, b1, 0, 0, xv-1, yv-1, xv, yv);
  FillSection(b2, c, b1, MX(b1)-1, 0, xv, yv-1, xv, yv);
  FillSection(b2, c, b1, 0, MY(b1)-1, xv-1, yv, xv, yv);
  FillSection(b2, c, b1, MX(b1)-1, MY(b1)-1, xv, yv, xv, yv);

  // Draw the front edges of the blocks, where not adjacent to other blocks.

  for (y = 0; y <= MY(b1); y++)
    for (x = 0; x <= MX(b1); x++) {
      if (Mg(b1, x, y) != Mg(b1, x, y - 1)) {
        if (!fColor)
          LineX(b2, x*ds.horsiz, (x+1)*ds.horsiz, y*ds.versiz, fOff);
        else
          LineXC(c, x*ds.horsiz, (x+1)*ds.horsiz, y*ds.versiz, ds.kvTrim);
      }
      if (Mg(b1, x, y) != Mg(b1, x - 1, y)) {
        if (!fColor)
          LineY(b2, x*ds.horsiz, y*ds.versiz, (y+1)*ds.versiz, fOff);
        else
          LineYC(c, x*ds.horsiz, y*ds.versiz, (y+1)*ds.versiz, ds.kvTrim);
      }

      // Draw visible edges of the block that go to the vanishing point. Don't
      // draw lines if the block is solid and the line would be behind it.

      if (FCorner(b1, x, y))
        PLine(b2, c, b1, x*ds.horsiz, y*ds.versiz, ds.horv, ds.verv, xv, yv);
    }
}


// Replace a bitmap with a single point perspective graphic of it. Implements
// the Render Pyramid command.

flag DrawPyramid(bitmap *pb, bitmap *pc)
{
  bitmap b = *pb, b1 = NULL;
  int x, y;
  flag fColor = (pc != NULL);

  x = MX(b)*ds.horsiz + 1; y =  MY(b)*ds.versiz + 1;
  if (FErrorRange("X cell size", ds.horsiz, 2, (xBitmap-1) / MX(b)))
    return fFalse;
  if (FErrorRange("Y cell size", ds.versiz, 2, (yBitmap-1) / MY(b)))
    return fFalse;
  if (FErrorRange("X vanishing point", ds.horv, 0, x-1))
    return fFalse;
  if (FErrorRange("Y vanishing point", ds.verv, 0, y-1))
    return fFalse;
  if (!fColor) {
    if (!FBitmapCopy(b, &b1))
      return fFalse;
    if (FBitmapSizeSet(pb, x, y)) {
      BitmapSet(*pb, !ds.fGroundShade);
      UpdateDisplay();
      FillPyramid(*pb, NULL, b1);
    }
    DeallocateBitmap(b1);
  } else {
    if (FColmapSizeSet(pc, x, y)) {
      ColmapSet(*pc, kvWhite);
      UpdateDisplay();
      FillPyramid(NULL, *pc, b);
    }
  }
  return fTrue;
}


/*
******************************************************************************
** File Reading Routines
******************************************************************************
*/

// Load a Daedalus wireframe file, into a new list of lines.

long ReadWirelist(COOR **pcoor, FILE *file)
{
  COOR *coor;
  long count = 0, i;
  real x1, y1, z1, x2, y2, z2;
  char ch, ch2;

  ch = getbyte(); ch2 = getbyte();
  if (ch != 'D' || ch2 != 'W') {
    PrintSz_W("This file does not look like a Daedalus wireframe.\n");
    return -1;
  }

  // Figure out how many coordinate pairs are in this wireframe file.
  ch = getbyte();
  if (ch == '#')
    fscanf(file, "%ld", &count);
  else {
    i = ftell(file);
    loop {
      ReadCoordinates(file, &x1, &y1, &z1, &x2, &y2, &z2);
      count++;
      if (fscanf(file, "%c", S1(&ch)) < 0) {
        count--;
        break;
      }
    }
    fseek(file, i, SEEK_SET);
  }

  // Allocate a line list of the appropriate size.
  coor = RgAllocate(count, COOR);
  if (coor == NULL)
    return -1;
  if (*pcoor != NULL)
    DeallocateP(*pcoor);
  *pcoor = coor;

  // Load the appropriate number of actual lines from the file.
  for (i = 0; i < count; i++) {
    ReadCoordinates(file, &x1, &y1, &z1, &x2, &y2, &z2);
    coor[i].x1 = x1; coor[i].y1 = y1; coor[i].z1 = z1;
    coor[i].x2 = x2; coor[i].y2 = y2; coor[i].z2 = z2;
  }
  return count;
}


// Load a Daedalus patch file, into a new list of patches.

long ReadPatchlist(PATCH **ppatch, FILE *file)
{
  PATCH *patch;
  long count = 0, i;
  int nTrans = ds.nTrans, status, nT;
  real x1, y1, z1, x2, y2, z2, x3, y3, z3;
  KV kv = ds.kvObject;
  char ch, ch2;

  ch = getbyte(); ch2 = getbyte();
  if (ch != 'D' || ch2 != 'P') {
    PrintSz_W("This file does not look like a Daedalus patch file.\n");
    return -1;
  }

  // Figure out how many patches are in this patch file.
  ch = getbyte();
  if (ch == '#')
    fscanf(file, "%ld", &count);
  else {
    i = ftell(file);
    loop {
      ReadPatch1(file, &status);
      switch (status) {
      case -1: case 8: ReadColor(file); break;
      case -2: ReadPatch1(file, &nT); break;
      default:
        count++;
        ReadPatch2(file, &x1, &y1, &z1, &x2, &y2, &z2, &x3, &y3, &z3);
        if (status & 32)
          ReadPatch3(file, &x1, &y1, &z1);
      }
      if (fscanf(file, "%c", S1(&ch)) < 0) {
        count--;
        break;
      }
    }
    fseek(file, i, SEEK_SET);
  }

  // Allocate a patch list of the appropriate size.
  patch = RgAllocate(count, PATCH);
  if (patch == NULL)
    return -1;
  if (*ppatch != NULL)
    DeallocateP(*ppatch);
  *ppatch = patch;

  // Load the appropriate number of actual patches from the file.
  for (i = 0; i < count; ) {
    ReadPatch1(file, &status);
    switch (status) {
    case -1: case 8: kv = ReadColor(file); break;
    case -2: ReadPatch1(file, &nTrans); break;
    default:
      ReadPatch2(file, &x1, &y1, &z1, &x2, &y2, &z2, &x3, &y3, &z3);
      patch[i].p[0].x = x1;
      patch[i].p[0].y = y1;
      patch[i].p[0].z = z1;
      patch[i].p[1].x = x2;
      patch[i].p[1].y = y2;
      patch[i].p[1].z = z2;
      patch[i].p[2].x = x3;
      patch[i].p[2].y = y3;
      patch[i].p[2].z = z3;
      patch[i].p[0].fLine = (status & 4) > 0;
      patch[i].p[1].fLine = (status & 2) > 0;
      patch[i].p[2].fLine = FOdd(status);

      // Load the fourth point if this is a quadrilateral patch.
      if (status & 32) {
        ReadPatch3(file, &x1, &y1, &z1);
        patch[i].p[3].x = x1;
        patch[i].p[3].y = y1;
        patch[i].p[3].z = z1;
        patch[i].p[3].fLine = (status & 16) > 0;
        patch[i].cpt = 4;
      } else {
        patch[i].p[3].x = patch[i].p[3].y = patch[i].p[3].z = 0.0;
        patch[i].p[3].fLine = fFalse;
        patch[i].cpt = 3;
      }
      patch[i].kv = kv;
      patch[i].nTrans = nTrans;
      i++;
    }
  }
  return count;
}


/*
******************************************************************************
** File Writing Routines
******************************************************************************
*/

// Save a list of lines to a Daedalus wireframe file.

void WriteWireframe(FILE *file, CONST COOR *coor, long ccoor)
{
  long i;

  fprintf(file, "DW#\n%ld\n", ccoor);
  for (i = 0; i < ccoor; i++) {
    fprintf(file, "%d %d %d %d %d %d\n",
      (int)coor[i].x1, (int)coor[i].y1, (int)coor[i].z1,
      (int)coor[i].x2, (int)coor[i].y2, (int)coor[i].z2);
  }
}


// Save a list of patches to a Daedalus patch file.

void WritePatches(FILE *file, CONST PATCH *patch, long cpatch)
{
  char sz[cchSzDef];
  CONST PATCH *pat;
  long i;
  int nTransLast = 0, j;
  KV kvPatchLast = ds.kvObject;

  fprintf(file, "DP#\n%ld\n", cpatch);
  for (i = 0; i < cpatch; i++) {
    pat = &patch[i];

    // Save color or transparency change if this patch differs from previous.
    if (pat->kv != kvPatchLast) {
      kvPatchLast = pat->kv;
      ConvertKvToSz(pat->kv, sz, cchSzDef);
      fprintf(file, "-1 %s\n", sz);
    }
    if (pat->nTrans != nTransLast) {
      nTransLast = pat->nTrans;
      fprintf(file, "-2 %d\n", nTransLast);
    }

    // Save the actual patch points.
    Assert(pat->cpt == 3 || pat->cpt == 4);
    fprintf(file, "%d", (pat->cpt == 4 ? 32 : 0) |
      pat->p[0].fLine*4 | pat->p[1].fLine*2 | pat->p[2].fLine |
      pat->p[3].fLine*16);
    for (j = 0; j < pat->cpt; j++)
      fprintf(file, " %d %d %d",
        (int)pat->p[j].x, (int)pat->p[j].y, (int)pat->p[j].z);
    fprintf(file, "\n");
  }
}


// Append a line to a coordinate list. Called by the functions that create
// wireframes in memory.

void WriteCoordinates(COOR *coor, int x1, int y1, int z1,
  int x2, int y2, int z2)
{
  COOR *pcoor;

  if (coor != NULL) {
    pcoor = &coor[ds.cCoorPatch];
    pcoor->x1 = x1; pcoor->y1 = y1; pcoor->z1 = z1;
    pcoor->x2 = x2; pcoor->y2 = y2; pcoor->z2 = z2;
  }
  ds.cCoorPatch++;
}


// Given a pair of coordinates assumed to form opposite corners of a
// rectangle of a given size in a grid of such rectangles, adjust them so even
// rows and columns will have the specified adjusted size.

void BiasCoordinates(int x, int y, int horsiz, int versiz,
  int horbias, int verbias, int *x1, int *y1, int *x2, int *y2)
{
  if (FOdd(x))
    *x1 += horbias - horsiz;
  else
    *x2 += horbias - horsiz;
  if (FOdd(y))
    *y1 += verbias - versiz;
  else
    *y2 += verbias - versiz;
}


// Append a quadrilateral patch to a patch list. Called by the functions that
// create patches in memory.

void WritePatch(PATCH *patch, PATN pat[cPatch], flag fReflect, KV kv)
{
  PATCH *ppatch;
  int i, j;

  if (patch != NULL) {
    ppatch = &patch[ds.cCoorPatch];
    ppatch->cpt = 4;
    ppatch->kv = kv;
    ppatch->nTrans = ds.nTrans;
    for (i = 0; i < 4; i++) {
      j = fReflect ? i : 4-i & 3;
      ppatch->p[i].x = pat[j].x;
      ppatch->p[i].y = pat[j].y;
      ppatch->p[i].z = pat[j].z;
      j = fReflect ? i : i^3;
      ppatch->p[i].fLine = pat[j].fLine;
    }
  }
  ds.cCoorPatch++;
}


/*
******************************************************************************
** Create Wireframe
******************************************************************************
*/

// Generate lines in a plane, forming rectangles around on pixels in a bitmap.
// Called from CreateOverviewLine().

void WriteLineLevel(COOR *coor, constbitmap b, int height)
{
  int x, y, x1, y1, x2, y2, count;

  // For each point, generate up to two lines, one for each axis.
  for (x = 0; x <= MX(b); x++)
    for (y = 0; y <= MY(b); y++) {
      count = Mg(b, x, y - 1) + Mg(b, x, y);
      if (count == 1 || count == 2 - ds.fMerge) {
        x1 = ds.hormin + ds.horsiz * x;
        y1 = ds.vermin + ds.versiz * y;
        x2 = x1 + ds.horsiz;
        y2 = y1;
        BiasCoordinates(x, y, ds.horsiz, ds.versiz, ds.horbias, ds.verbias,
          &x1, &y1, &x2, &y2);
        if (x1 != x2)
          WriteCoordinates(coor, x1, y1, height, x2, y1, height);
      }
      count = Mg(b, x - 1, y) + Mg(b, x, y);
      if (count == 1 || count == 2 - ds.fMerge) {
        x1 = ds.hormin + ds.horsiz * x;
        y1 = ds.vermin + ds.versiz * y;
        x2 = x1;
        y2 = y1 + ds.versiz;
        BiasCoordinates(x, y, ds.horsiz, ds.versiz, ds.horbias, ds.verbias,
          &x1, &y1, &x2, &y2);
        if (y1 != y2)
          WriteCoordinates(coor, x1, y1, height, x1, y2, height);
      }
    }
}


// Generate a wireframe forming blocks around pixels in a bitmap. Called by
// CreateWireframe for 2D bitmaps.

flag CreateOverviewLine(constbitmap b, COOR *coor)
{
  int x, y, x1, y1, x2, y2, count;

  if (FErrorRange("X block size", ds.horsiz, 1, 1000))
    return fFalse;
  if (FErrorRange("Y block size", ds.versiz, 1, 1000))
    return fFalse;
  if (FErrorRange("Block height", ds.depsiz, 0, 1000))
    return fFalse;
  if (FErrorRange("X thickness bias", ds.horbias, 0, ds.horsiz*2 - 1))
    return fFalse;
  if (FErrorRange("Y thickness bias", ds.verbias, 0, ds.versiz*2 - 1))
    return fFalse;

  // Generate lines for the top and bottom of blocks.
  WriteLineLevel(coor, b, ds.depmin);
  if (ds.depsiz == 0)
    return fTrue;
  WriteLineLevel(coor, b, ds.depmin + ds.depsiz);

  // Generate lines through the z-axis from the top to bottom of blocks.
  for (y = 0; y <= MY(b); y++)
    for (x = 0; x <= MX(b); x++) {
      count = Mg(b, x - 1, y - 1) + Mg(b, x, y - 1) +
        Mg(b, x - 1, y) + Mg(b, x, y);
      if (FOdd(count) ||
        (count == 2 && Mg(b, x - 1, y - 1) == Mg(b, x, y)) ||
        (!ds.fMerge && count > 0)) {
        x1 = ds.hormin + ds.horsiz*x;
        y1 = ds.vermin + ds.versiz*y;
        x2 = x1;
        y2 = y1;
        BiasCoordinates(x, y, ds.horsiz, ds.versiz, ds.horbias, ds.verbias,
          &x1, &y1, &x2, &y2);
        WriteCoordinates(coor,
          x1, y1, ds.depmin, x1, y1, ds.depmin + ds.depsiz);
      }
    }
  return fTrue;
}


// Generate a wireframe forming blocks around pixels in a 3D bitmap. Called by
// CreateWireframe for 3D bitmaps.

flag CreateCubicLine(constcubemap t, COOR *coor)
{
  int x, y, z, x1, y1, z1, count;
  flag fTest;

  if (FErrorRange("Block length", ds.horsiz, 1, 1000))
    return fFalse;
  if (FErrorRange("Block width", ds.versiz, 1, 1000))
    return fFalse;
  if (FErrorRange("Block depth", ds.depsiz, 1, 1000))
    return fFalse;

  // For each point, generate up to three lines, one for each axis.
  for (y = 0; y <= TX(t); y++)
    for (z = 0; z <= TZ(t); z++)
      for (x = 0; x <= TY(t); x++) {
        x1 = ds.hormin + ds.horsiz*x;
        y1 = ds.vermin + ds.versiz*(TY(t)-y);
        z1 = ds.depmin + ds.depsiz*(TZ(t)-z);
        count = Tg(t, x - 1, y - 1, z) + Tg(t, x, y - 1, z) +
          Tg(t, x - 1, y, z) + Tg(t, x, y, z);
        fTest = Tg(t, x - 1, y - 1, z) == Tg(t, x, y, z);
        if (FOdd(count) || (count == 2 && fTest) || (!ds.fMerge && count > 0))
          WriteCoordinates(coor, x1, y1, z1, x1, y1, z1 - ds.depsiz);
        count = Tg(t, x - 1, y, z - 1) + Tg(t, x, y, z - 1) +
          Tg(t, x - 1, y, z) + Tg(t, x, y, z);
        fTest = Tg(t, x - 1, y, z - 1) == Tg(t, x, y, z);
        if (FOdd(count) || (count == 2 && fTest) || (!ds.fMerge && count > 0))
          WriteCoordinates(coor, x1, y1, z1, x1, y1 - ds.versiz, z1);
        count = Tg(t, x, y - 1, z - 1) + Tg(t, x, y, z - 1) +
          Tg(t, x, y - 1, z) + Tg(t, x, y, z);
        fTest = Tg(t, x, y - 1, z - 1) == Tg(t, x, y, z);
        if (FOdd(count) || (count == 2 && fTest) || (!ds.fMerge && count > 0))
          WriteCoordinates(coor, x1, y1, z1, x1 + ds.horsiz, y1, z1);
      }
  return fTrue;
}


// Create a wireframe forming blocks around pixels in a bitmap, in a new line
// list. Implements the Make Wireframe Bitmap Overview command.

long CreateWireframe(constbitmap b, COOR **pcoor, flag f3D)
{
  COOR *coor;

  // Figure out how many lines will be in the wireframe.
  ds.cCoorPatch = 0;
  if (!f3D)
    CreateOverviewLine(b, NULL);
  else
    CreateCubicLine(b, NULL);

  // Create a line list of appropriate size.
  coor = RgAllocate(ds.cCoorPatch, COOR);
  if (coor == NULL)
    return -1;
  if (*pcoor != NULL)
    DeallocateP(*pcoor);
  *pcoor = coor;

  // Actually generate the wireframe.
  ds.cCoorPatch = 0;
  if (!f3D)
    CreateOverviewLine(b, coor);
  else
    CreateCubicLine(b, coor);
  return ds.cCoorPatch;
}


/*
******************************************************************************
** Create Patches
******************************************************************************
*/

// Figure out the color to use for a patch.
#define KvObjectColor(c, x, y) \
  ((c) != NULL && FLegalC(c, x, y) ? CG(c, x, y) : ds.kvObject)
#define KvObjectColor3(c, x, y, z, t) \
  KvObjectColor(c, Tx(t, x, z), Ty(t, y, z))

// Generate patches in a plane, forming rectangles corresponding to on pixels
// in a bitmap. Called from CreateOverviewPatch().

void WritePatchLevel(PATCH *patch, constbitmap b, int height, flag fReflect,
  constcolmap c)
{
  int x, y, x1, y1, x2, y2;
  PATN pat[cPatch];
  KV kv;

  pat[0].z = pat[1].z = pat[2].z = pat[3].z = height;
  for (y = 0; y < MY(b); y++)
    for (x = 0; x < MX(b); x++)
      if (Mg(b, x, y)) {

        // For each on pixel, generate a rectangular patch for it.
        kv = KvObjectColor(c, x, y);
        x1 = ds.hormin + ds.horsiz * x;
        y1 = ds.vermin + ds.versiz * y;
        x2 = x1 + ds.horsiz;
        y2 = y1 + ds.versiz;
        BiasCoordinates(x, y, ds.horsiz, ds.versiz, ds.horbias, ds.verbias,
          &x1, &y1, &x2, &y2);
        pat[0].x = x1; pat[0].y = y1;
        pat[1].x = x1; pat[1].y = y2;
        pat[2].x = x2; pat[2].y = y2;
        pat[3].x = x2; pat[3].y = y1;
        pat[0].fLine = !ds.fMerge || !Mg(b, x-1, y);
        pat[1].fLine = !ds.fMerge || !Mg(b, x, y+1);
        pat[2].fLine = !ds.fMerge || !Mg(b, x+1, y);
        pat[3].fLine = !ds.fMerge || !Mg(b, x, y-1);
        WritePatch(patch, pat, fReflect, kv);
      }
}


// Generate patches forming blocks around pixels in a bitmap. Called by
// CreatePatches for 2D bitmaps.

void CreateOverviewPatch(constbitmap b, PATCH *patch, constcolmap c)
{
  int x, y, x1, y1, x2, y2;
  PATN pat[cPatch];
  KV kv;

  // Generate patches for the top and bottom of blocks.
  if (ds.horbias != 0 && ds.verbias != 0)
    WritePatchLevel(patch, b, ds.depmin, fFalse, c);
  if (ds.depsiz == 0)
    return;
  if (ds.horbias != 0 && ds.verbias != 0)
    WritePatchLevel(patch, b, ds.depmin + ds.depsiz, fTrue, c);

  // Generate patches through the z-axis from the top to bottom of blocks.
  pat[0].z = pat[3].z = ds.depmin + ds.depsiz;
  pat[1].z = pat[2].z = ds.depmin;
  pat[1].fLine = pat[3].fLine = fTrue;
  for (y = 0; y <= MY(b); y++)
    for (x = 0; x <= MX(b); x++) {
      x1 = ds.hormin + ds.horsiz*x;
      y1 = ds.vermin + ds.versiz*y;
      x2 = x1 + ds.horsiz;
      y2 = y1 + ds.versiz;
      BiasCoordinates(x, y, ds.horsiz, ds.versiz, ds.horbias, ds.verbias,
        &x1, &y1, &x2, &y2);

      // For each point, generate up to two patches, one for the face
      // perpendicular to each axis.
      if (Mg(b, x, y) != Mg(b, x, y-1) && x1 != x2) {
        kv = KvObjectColor(c, x, y - Mg(b, x, y-1));
        pat[0].x = pat[1].x = x1;
        pat[2].x = pat[3].x = x2;
        pat[0].y = pat[1].y = pat[2].y = pat[3].y = y1;
        pat[0].fLine = !ds.fMerge || Mg(b, x, y) == Mg(b, x-1, y-1) ||
          Mg(b, x, y-1) == Mg(b, x-1, y);
        pat[2].fLine = !ds.fMerge || Mg(b, x, y) == Mg(b, x+1, y-1) ||
          Mg(b, x, y-1) == Mg(b, x+1, y);
        WritePatch(patch, pat, Mg(b, x, y-1), kv);
      }
      if (Mg(b, x, y) != Mg(b, x-1, y) && y1 != y2) {
        kv = KvObjectColor(c, x - Mg(b, x-1, y), y);
        pat[0].x = pat[1].x = pat[2].x = pat[3].x = x1;
        pat[0].y = pat[1].y = y1;
        pat[2].y = pat[3].y = y2;
        pat[0].fLine = !ds.fMerge || Mg(b, x, y) == Mg(b, x-1, y-1) ||
          Mg(b, x-1, y) == Mg(b, x, y-1);
        pat[2].fLine = !ds.fMerge || Mg(b, x, y) == Mg(b, x-1, y+1) ||
          Mg(b, x-1, y) == Mg(b, x, y+1);
        WritePatch(patch, pat, Mg(b, x, y), kv);
      }
    }
}


// Generate patches forming blocks around pixels in a 3D bitmap. Called by
// CreatePatches for 3D bitmaps.

void CreateCubicPatch(constcubemap t, PATCH *patch, constcolmap c)
{
  int x, y, z, x1, y1, z1, x2, y2, z2, i;
  PATN pat[cPatch];
  KV kv;

  for (y = 0; y <= TY(t); y++) {
    y1 = ds.vermin + ds.versiz*(TY(t)-y);
    y2 = y1 - ds.versiz;
    for (z = 0; z <= TZ(t); z++) {
      z1 = ds.depmin + ds.depsiz*(TZ(t)-z);
      z2 = z1 - ds.depsiz;
      for (x = 0; x <= TX(t); x++) {
        x1 = ds.hormin + ds.horsiz*x;
        x2 = x1 + ds.horsiz;

        // For each point, generate up to three patches, one for the face
        // perpendicular to each axis.
        if (Tg(t, x, y, z) != Tg(t, x, y, z - 1)) {
          kv = KvObjectColor3(c, x, y, z - Tg(t, x, y, z - 1), t);
          for (i = 0; i < 4; i++) {
            pat[i].x = x1; pat[i].y = y1; pat[i].z = z1;
          }
          pat[2].x = pat[3].x = x2;
          pat[1].y = pat[2].y = y2;
          pat[0].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x - 1, y, z) ||
            Tg(t, x, y, z - 1) != Tg(t, x - 1, y, z - 1);
          pat[1].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x, y + 1, z) ||
            Tg(t, x, y, z - 1) != Tg(t, x, y + 1, z - 1);
          pat[2].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x + 1, y, z) ||
            Tg(t, x, y, z - 1) != Tg(t, x + 1, y, z - 1);
          pat[3].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x, y - 1, z) ||
            Tg(t, x, y, z - 1) != Tg(t, x, y - 1, z - 1);
          WritePatch(patch, pat, Tg(t, x, y, z - 1), kv);
        }
        if (Tg(t, x, y, z) != Tg(t, x - 1, y, z)) {
          kv = KvObjectColor3(c, x - Tg(t, x - 1, y, z), y, z, t);
          for (i = 0; i < 4; i++) {
            pat[i].x = x1; pat[i].y = y1; pat[i].z = z1;
          }
          pat[2].y = pat[3].y = y2;
          pat[1].z = pat[2].z = z2;
          pat[0].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x, y - 1, z) ||
            Tg(t, x - 1, y, z) != Tg(t, x - 1, y - 1, z);
          pat[1].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x, y, z + 1) ||
            Tg(t, x - 1, y, z) != Tg(t, x - 1, y, z + 1);
          pat[2].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x, y + 1, z) ||
            Tg(t, x - 1, y, z) != Tg(t, x - 1, y + 1, z);
          pat[3].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x, y, z - 1) ||
            Tg(t, x - 1, y, z) != Tg(t, x - 1, y, z - 1);
          WritePatch(patch, pat, Tg(t, x - 1, y, z), kv);
        }
        if (Tg(t, x, y, z) != Tg(t, x, y - 1, z)) {
          kv = KvObjectColor3(c, x, y - Tg(t, x, y - 1, z), z, t);
          for (i = 0; i < 4; i++) {
            pat[i].x = x1; pat[i].y = y1; pat[i].z = z1;
          }
          pat[2].x = pat[3].x = x2;
          pat[1].z = pat[2].z = z2;
          pat[0].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x - 1, y, z) ||
            Tg(t, x, y - 1, z) != Tg(t, x - 1, y - 1, z);
          pat[1].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x, y, z + 1) ||
            Tg(t, x, y - 1, z) != Tg(t, x, y - 1, z + 1);
          pat[2].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x + 1, y, z) ||
            Tg(t, x, y - 1, z) != Tg(t, x + 1, y - 1, z);
          pat[3].fLine = !ds.fMerge || Tg(t, x, y, z) != Tg(t, x, y, z - 1) ||
            Tg(t, x, y - 1, z) != Tg(t, x, y - 1, z - 1);
          WritePatch(patch, pat, Tg(t, x, y, z), kv);
        }
      }
    }
  }
}


// Create patches forming blocks around pixels in a bitmap, in a new patch
// list. Implements the Make Patch Bitmap Overview command.

long CreatePatches(constbitmap b, PATCH **ppatch, flag f3D, constcolmap c)
{
  PATCH *patch;

  // Figure out how many patches will be in the patch list.
  ds.cCoorPatch = 0;
  if (!f3D)
    CreateOverviewPatch(b, NULL, c);
  else
    CreateCubicPatch(b, NULL, c);

  // Create a patch list of appropriate size.
  patch = RgAllocate(ds.cCoorPatch, PATCH);
  if (patch == NULL)
    return -1;
  if (*ppatch != NULL)
    DeallocateP(*ppatch);
  *ppatch = patch;

  // Actually generate the patches.
  ds.cCoorPatch = 0;
  if (!f3D)
    CreateOverviewPatch(b, patch, c);
  else
    CreateCubicPatch(b, patch, c);
  return ds.cCoorPatch;
}

/* draw.cpp */
