/*
** Daedalus (Version 2.5) File: graphics.h
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains definitions for general graphics routines, that operate
** on a monochrome bitmap, unrelated to Mazes.
**
** Created: 7/1/1993.
** Last code change: 10/10/2013.
*/

/*
******************************************************************************
** Constants
******************************************************************************
*/

#define xBitmap (0x7FFFFFFF - 31)
#define yBitmap 0x7FFFFFFF
#define xStart 64
#define yStart 64
#define zBitmapStack 1024
#define wClient1Def 16
#define wClient2Def 16
#define wClient3Def 16
#define wClient4Def 4

#define DIRS 4
#define DIRS1 (DIRS-1)
#define DIRS2 8


/*
******************************************************************************
** Macros
******************************************************************************
*/

#define DeallocateBitmap(b) DeallocateP(b)
#define CbBitmapRow(x) ((((x) + 31) >> 5) << 2)
#define CbBitmap(x, y) LMul(y, CbBitmapRow(x))
#define MX(b) ((b)->x)
#define MY(b) ((b)->y)
#define MI(b, x, y) ((long)(y)*((b)->clRow) + ((x) >> 5))
#define MJ(b, i) ((dword *)((byte *)(b) + cbBH + ((i) << 2)))
#define MA(b, x, y) (*MJ(b, MI(b, x, y)))
#define MB(x) (1L << ((x)&31 ^ 7))
#define MGCore(b, x, y) ((MA(b, x, y) & MB(x)) != 0)
#define MG(b, x, y) _MG(b, x, y)
#define MP(b, x, y, o) _MP(b, x, y, o)
#define CopyClient(b1, b2) \
  (b2)->wClient1 = (b1)->wClient1; (b2)->wClient2 = (b1)->wClient2; \
  (b2)->wClient3 = (b1)->wClient3; (b2)->wClient4 = (b1)->wClient4;

#define DirInc(d) d = ((d) + 1) & DIRS1
#define DirDec(d) d = ((d) - 1) & DIRS1
#define FDitherCore(col, x, y) ((col) >= rgnDitherPoint[((y) << 2) + (x)])
#define FDither(col, x, y) FDitherCore(col, (x) & 3, (y) & 3)
#define NBinary(f0, f1, f2, f3, f4, f5, f6, f7) \
  ((f0 << 4) | (f1 << 1) | (f2 << 5) | (f3 << 2) | \
  (f4 << 6) | (f5 << 3) | (f6 << 7) | (f7 << 0))


/*
******************************************************************************
** Types
******************************************************************************
*/

typedef long KV;

typedef struct _bitmapheader {
  int x;         // Horizontal size in pixels
  int y;         // Vertical size in pixels
  int clRow;     // Longs per horizontal row
  int fColor;    // Whether bitmap is color
  int wClient1;  // 3D bitmap X size
  int wClient2;  // 3D bitmap Y size
  int wClient3;  // 3D bitmap Z size
  int wClient4;  // 3D bitmap W size
  dword bits[1]; // Actual bitmap pixels
} BH;

#define cbBH (sizeof(BH)-sizeof(dword))

typedef BH *bitmap;
typedef BH CONST *constbitmap;

typedef struct _graphicssettings {
  flag fTraceDot;
  flag fErrorCheck;

  flag fSlowDot;
  bitmap *pbFocus;
  int turtlex;
  int turtley;
  int turtles;
  int turtlea;
  flag turtlec;
  flag turtleo;
  int turtlet;
  int grfLifeDie;
  int grfLifeBorn;
} GS;

extern CONST char *rgszDir[DIRS];
extern CONST int xoff[DIRS2], yoff[DIRS2], xoff2[DIRS2], yoff2[DIRS2];
extern CONST int rgfOr[3], rgfAnd[3], rgfXor[3];
extern CONST int rgnDitherPoint[16];
extern CONST int rggrfNeighbor[25];
extern GS gs;


/*
******************************************************************************
** Bitmap Primitives
******************************************************************************
*/

extern void UpdateDisplay(void);
extern void ScreenDot(int, int, bit, KV);

extern flag FAllocateBitmap(bitmap *, int, int, constbitmap);
#define FSlowDot() (gs.fTraceDot || gs.fErrorCheck)
extern flag FValidate(constbitmap, int, int, CONST char *);
extern void M0(bitmap, int, int);
extern void M1(bitmap, int, int);
extern void MR(bitmap, int, int);
extern bit _MG(constbitmap, int, int);
extern void _MP(bitmap, int, int, bit);
extern bit Mg(constbitmap, int, int);
#define MgFast(b, x, y) (FLegal(b, x, y) ? MGCore(b, x, y) : fOff)
extern void Mp(bitmap, int, int, bit);
#define FNull(b) (MX(b) < 1 || MY(b) < 1)
#define F64K(b) (MX(b) > 0xFFFF || MY(b) > 0xFFFF)
#define FLegal(b, x, y) \
  ((x) >= 0 && (x) < MX(b) && (y) >= 0 && (y) < MY(b))
#define FLegalOff(b, x, y) (FLegal(b, x, y) && !MG(b, x, y))
extern void Legalize(constbitmap, int *, int *);


/*
******************************************************************************
** Bitmap Block Editing
******************************************************************************
*/

extern void Block(bitmap, int, int, int, int, bit);
extern void BlockReverse(bitmap, int, int, int, int);
extern void Box(bitmap, int, int, int, int, int, int, bit);
#define Edge(b, x1, y1, x2, y2, o) Box(b, x1, y1, x2, y2, 1, 1, o)
#define BoxAll(b, xsiz, ysiz, o) \
  Box(b, 0, 0, MX(b)-1, MY(b)-1, xsiz, ysiz, o)
#define EdgeAll(b, o) BoxAll(b, 1, 1, o)
extern void BlockMove(constbitmap, bitmap, int, int, int, int, int, int);
extern void BlockMoveOrAnd(constbitmap, bitmap, int, int, int, int, int, int,
  CONST int *);
#define BlockPutOrAnd(b1, b2, x, y, rgf) \
  BlockMoveOrAnd(b1, b2, 0, 0, MX(b1)-1, MY(b1)-1, x, y, rgf)


/*
******************************************************************************
** Bitmap Line Drawing
******************************************************************************
*/

extern void LineX(bitmap, int, int, int, bit);
extern void LineY(bitmap, int, int, int, bit);
extern void LineZ(bitmap, int, int, int, int, bit);
extern void Line(bitmap, int, int, int, int, bit);
extern flag FLineUntil(bitmap, int, int, int, int, int *, int *,
  bit, flag, flag);
extern void LineSetup(constbitmap, MM *, int, int, int, int);
extern flag FTriangle(bitmap, int, int, int, int, int, int, int);
extern flag FQuadrilateral(bitmap, int, int, int, int, int, int, int, int,
  int);
extern flag Turtle(bitmap, CONST char *);


/*
******************************************************************************
** Bitmap Curve Drawing
******************************************************************************
*/

extern void Disk(bitmap, int, int, int, int, bit);
extern void Circle(bitmap, int, int, int, int, bit);
extern void ArcQuadrant(bitmap, int, int, int, int, bit);
extern void ArcQuadrantSub(bitmap, int, int, int, int,
  int, int, int *, int *, bit);
extern void ArcPolygon(bitmap, int, int, int, int, int, int, int, int, bit);
#define Polygon(b, x, y, rx, ry, d, o) \
  ArcPolygon(b, x, y, rx, ry, 0, d, 1, d, o)
extern void ArcReal(bitmap, real, real, real, real, real, bit);


/*
******************************************************************************
** Bitmap Flooding
******************************************************************************
*/

#define FLegalFill(b, x, y, o) \
  (FLegal(b, x, y) && MGCore(b, x, y) != (o))
#define FillPush(xnew, ynew) rgpt[ipt].x = (xnew); rgpt[ipt].y = (ynew); ipt++
#define FillPop(xnew, ynew) ipt--; xnew = rgpt[ipt].x; ynew = rgpt[ipt].y
extern flag FFillCore(bitmap, int, int, bit, flag);
#define FFill(b, x, y, o) FFillCore(b, x, y, o, fFalse)
#define FFillReverse(b, x, y) FFill(b, x, y, !Mg(b, x, y))
#define FFlood(b, x, y, o) FFillCore(b, x, y, o, fTrue)
#define FFloodReverse(b, x, y) FFlood(b, x, y, !Mg(b, x, y))
extern long AnalyzePerimeter(constbitmap, int, int, flag);


/*
******************************************************************************
** Bitmap Transformations
******************************************************************************
*/

extern long BitmapCount(constbitmap);
extern flag FBitmapFind(constbitmap, int *, int *, bit);
extern flag FBitmapSet(bitmap, bit);
extern void BitmapSet(bitmap, bit);
#define BitmapOff(b) BitmapSet(b, fOff)
#define BitmapOn(b) BitmapSet(b, fOn)
extern void BitmapReverse(bitmap);
extern void BitmapRandom(bitmap, int);
extern void BitmapMoire(bitmap);
extern flag FBitmapCopy(constbitmap, bitmap *);
extern void BitmapOr(bitmap, constbitmap);
extern void BitmapAnd(bitmap, constbitmap);
extern void BitmapXor(bitmap, constbitmap);
extern void BitmapBlend(bitmap, constbitmap);
extern void BitmapTessellate(bitmap, constbitmap);
extern flag FBitmapTessellateTo(bitmap *, int, int);
#define FBitmapTessellateBy(b, x, y) FBitmapTessellateTo(b, \
  (int)((real)MX(*b)*(x) + rMinute), (int)((real)MY(*b)*(y) + rMinute))
extern void BitmapFlipX(bitmap);
extern void BitmapFlipY(bitmap);
extern flag FBitmapTranspose(bitmap *);
#define BitmapRotateR(pb) FBitmapTranspose(pb); BitmapFlipX(*pb)
#define BitmapRotateL(pb) FBitmapTranspose(pb); BitmapFlipY(*pb)
#define BitmapRotateA(b) BitmapFlipX(b); BitmapFlipY(b)
extern flag FBitmapSizeSet(bitmap *, int, int);
extern flag FBitmapResizeTo(bitmap *, int, int);
#define FBitmapResizeBy(b, x, y) FBitmapResizeTo(b, MX(*b)+(x), MY(*b)+(y))
extern flag FBitmapShiftBy(bitmap *, int, int);
extern flag FBitmapZoomTo(bitmap *, int, int, flag);
#define FBitmapZoomBy(b, x, y, f) FBitmapZoomTo(b, \
  (int)((real)MX(*b)*(x) + rMinute), (int)((real)MY(*b)*(y) + rMinute), f)
extern flag FBitmapBias(bitmap *, int, int, int, int);
extern long BitmapSmooth(bitmap *, flag);
extern long BitmapSmoothCorner(bitmap, bit);
extern flag FBitmapCollapse(bitmap *);
extern long BitmapThicken(bitmap);
extern long BitmapThicken2(bitmap, int, flag);
extern long BitmapThinner(bitmap, flag);
extern flag FBitmapConvex(bitmap, int, int, int, int);
extern flag FBitmapAccentBoundary(bitmap *);
extern flag FBitmapAccentContrast(bitmap, flag);
extern flag FBitmapSlide(bitmap, int, int);
extern flag FBitmapEditRowColumn(bitmap *, int, int, flag, flag);
extern void BitmapSeen(bitmap, int, int, int);
extern flag FStereogram(bitmap, constbitmap, int, int);
extern long LifeGenerate(bitmap, flag);
extern flag FSetLife(CONST char *, int);


/*
******************************************************************************
** Bitmap File Routines
******************************************************************************
*/

extern flag FReadBitmapHeader(FILE *, flag, int *, int *, int *, int *);
extern flag FReadBitmapCore(bitmap *, FILE *, int, int);
extern void WriteBitmap(constbitmap, FILE *, KV, KV);
extern flag FReadText(bitmap *, FILE *);
extern void WriteText(constbitmap, FILE *, flag, flag, flag);
extern flag FReadXbm(bitmap *, FILE *);
extern void WriteXbm(constbitmap, FILE *, char *, char);
extern flag FReadDaedalusBitmapCore(bitmap *, FILE *, int, int);
extern void WriteDaedalusBitmap(constbitmap, FILE *, flag);

/* graphics.h */
